from pathlib import Path
from typing import Callable, Optional, Sequence, Union

import pandas

from beignet.transforms import Transform

from ._data_frame_dataset import DataFrameDataset


class ParquetDataset(DataFrameDataset):
    def __init__(
        self,
        root: Union[str, Path],
        path: Union[str, Path],
        *,
        columns: Optional[Sequence[str]],
        target_columns: Optional[Sequence[str]],
        transform: Union[Callable, Transform, None] = None,
        target_transform: Union[Callable, Transform, None] = None,
        **kwargs,
    ) -> None:
        """
        :param root: Root directory where the dataset subdirectory exists or,
            if :attr:`download` is ``True``, the directory where the dataset
            subdirectory will be created and the dataset downloaded.

        :param columns: x features of the dataset. items in the dataset are
            of the form ((columns), (target_columns)).

        :param target_columns: y features of the dataset. items in the dataset
            are of the form ((columns), (target_columns)).

        :param transform: A ``Callable`` or ``Transform`` that maps data to
            transformed data (default: ``None``).

        :param target_transform: ``Callable`` or ``Transform`` that maps a
            target to a transformed target (default: ``None``).
        """
        super().__init__(
            root,
            transform=transform,
            target_transform=target_transform,
        )

        self._path = path

        self._columns = columns

        self._target_columns = target_columns

        self._data = pandas.read_parquet(self._path, **kwargs)



import beignet
import torch


def test_chebyshev_polynomial_roots():
    torch.testing.assert_close(
        beignet.chebyshev_polynomial_roots(
            torch.tensor([1.0]),
        ),
        torch.tensor([]),
    )

    torch.testing.assert_close(
        beignet.chebyshev_polynomial_roots(
            torch.tensor([1.0, 2.0]),
        ),
        torch.tensor([-0.5]),
    )

    for i in range(2, 5):
        torch.testing.assert_close(
            beignet.trim_chebyshev_polynomial_coefficients(
                beignet.chebyshev_polynomial_roots(
                    beignet.chebyshev_polynomial_from_roots(
                        torch.linspace(-1, 1, i),
                    )
                ),
                tol=0.000001,
            ),
            beignet.trim_chebyshev_polynomial_coefficients(
                torch.linspace(-1, 1, i),
                tol=0.000001,
            ),
        )



import numpy
import torch

from ._evaluate_physicists_hermite_polynomial import (
    evaluate_physicists_hermite_polynomial,
)


def integrate_physicists_hermite_polynomial(
    input,
    order=1,
    k=None,
    lower_bound=0,
    scale=1,
    axis=0,
):
    if k is None:
        k = []

    input = torch.atleast_1d(input)

    lower_bound, scale = map(torch.tensor, (lower_bound, scale))

    if not numpy.iterable(k):
        k = [k]

    if len(k) > order:
        raise ValueError

    if lower_bound.ndim != 0:
        raise ValueError

    if scale.ndim != 0:
        raise ValueError

    if order == 0:
        return input

    input = torch.moveaxis(input, axis, 0)

    k = torch.tensor(list(k) + [0] * (order - len(k)))
    k = torch.atleast_1d(k)

    for i in range(order):
        n = input.shape[0]
        input *= scale
        tmp = torch.empty((n + 1,) + input.shape[1:], dtype=input.dtype)

        tmp[0] = input[0] * 0
        tmp[1] = input[0] / 2

        j = torch.arange(1, n)

        tmp[j + 1] = (input[j].T / (2 * (j + 1))).T

        tmp_value = evaluate_physicists_hermite_polynomial(lower_bound, tmp)
        tmp[0] += k[i] - tmp_value

        input = tmp

    input = torch.moveaxis(input, 0, axis)

    return input



from torch import Tensor

from ._quaternion_magnitude import quaternion_magnitude
from ._rotation_matrix_to_quaternion import (
    rotation_matrix_to_quaternion,
)


def rotation_matrix_magnitude(input: Tensor) -> Tensor:
    r"""
    Rotation matrix magnitudes.

    Parameters
    ----------
    input : Tensor, shape (..., 3, 3)
        Rotation matrices.

    Returns
    -------
    rotation_matrix_magnitudes: Tensor, shape (...)
        Angles in radians. Magnitudes will be in the range :math:`[0, \pi]`.
    """
    return quaternion_magnitude(
        rotation_matrix_to_quaternion(
            input,
        ),
    )



import beignet
import torch


def test_laguerre_polynomial_vandermonde():
    x = torch.arange(3)

    v = beignet.laguerre_polynomial_vandermonde(x, 3)

    assert v.shape == (3, 4)

    for i in range(4):
        torch.testing.assert_close(
            v[..., i],
            beignet.evaluate_laguerre_polynomial(
                x,
                torch.tensor([0.0] * i + [1.0]),
            ),
        )

    x = torch.tensor([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])

    v = beignet.laguerre_polynomial_vandermonde(x, 3)

    assert v.shape == (3, 2, 4)

    for i in range(4):
        torch.testing.assert_close(
            v[..., i],
            beignet.evaluate_laguerre_polynomial(
                x,
                torch.tensor([0.0] * i + [1.0]),
            ),
        )



import beignet
import torch


def test_evaluate_probabilists_hermite_polynomial_cartersian_3d():
    c1d = torch.tensor([4.0, 2.0, 3.0])
    c3d = torch.einsum("i,j,k->ijk", c1d, c1d, c1d)

    x = torch.rand(3, 5) * 2 - 1
    y = beignet.evaluate_polynomial(x, torch.tensor([1.0, 2.0, 3.0]))

    a, b, x3 = x
    y1, y2, y3 = y

    torch.testing.assert_close(
        beignet.evaluate_probabilists_hermite_polynomial_cartersian_3d(a, b, x3, c3d),
        torch.einsum("i,j,k->ijk", y1, y2, y3),
    )

    z = torch.ones([2, 3])
    res = beignet.evaluate_probabilists_hermite_polynomial_cartersian_3d(z, z, z, c3d)
    assert res.shape == (2, 3) * 3



import torch
from torch import Tensor


def evaluate_polynomial(
    input: Tensor, coefficients: Tensor, tensor: bool = True
) -> Tensor:
    r"""
    Parameters
    ----------
    input : Tensor

    coefficients : Tensor

    tensor : bool

    Returns
    -------
    output : Tensor
    """
    coefficients = torch.atleast_1d(coefficients)

    if tensor:
        coefficients = torch.reshape(
            coefficients,
            coefficients.shape + (1,) * input.ndim,
        )

    output = coefficients[-1] + torch.zeros_like(input)

    for i in range(2, coefficients.shape[0] + 1):
        output = coefficients[-i] + output * input

    return output



import math

import beignet
import torch


def test_evaluate_polynomial():
    output = beignet.evaluate_polynomial(
        torch.tensor([]),
        torch.tensor([1.0]),
    )

    assert math.prod(output.shape) == 0

    y = []

    input = torch.linspace(-1, 1, 50)

    for index in range(5):
        y = [
            *y,
            input**index,
        ]

    for index in range(5):
        torch.testing.assert_close(
            beignet.evaluate_polynomial(
                input,
                torch.tensor([0.0] * index + [1.0]),
            ),
            y[index],
        )

    torch.testing.assert_close(
        beignet.evaluate_polynomial(
            input,
            torch.tensor([0, -1, 0, 1]),
        ),
        input * (input**2 - 1),
    )

    for index in range(3):
        shape = (2,) * index

        input = torch.zeros(shape)

        output = beignet.evaluate_polynomial(
            input,
            torch.tensor([1.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_polynomial(
            input,
            torch.tensor([1.0, 0.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_polynomial(
            input,
            torch.tensor([1.0, 0.0, 0.0]),
        )

        assert output.shape == shape



from enum import Enum
from typing import Any, Callable, Dict, List, Tuple, Type, Union

import torch.utils._pytree
from torch import Tensor
from torch.nn import Module

from beignet.features import Feature


class Transform(Module):
    # Class attribute defining transformed types. Other types are
    # passed-through without any transformation
    #
    # We support both Types and callables that are able to do further checks
    # on the type of the input.
    _transformed_types: Tuple[Union[Type, Callable[[Any], bool]], ...] = (
        Tensor,
        str,
    )

    def __init__(self):
        super().__init__()

    def _check_inputs(self, inputs: List[Any]):
        """
        Parameters
        ----------
        inputs : List[Any]
            The inputs to be checked.
        """
        raise NotImplementedError

    def _get_params(self, inputs: List[Any]) -> Dict[str, Any]:
        """
        Parameters
        ----------
        inputs: List[Any]
            The list of input objects.

        Returns
        -------
        Dict[str, Any]
            A dictionary containing the parameters of the method.
        """
        return dict()

    def _transform(self, input: Any, parameters: Dict[str, Any]) -> Any:
        """
        Parameters
        ----------
        input : Any
            The input data to be transformed.

        parameters : Dict[str, Any]
            A dictionary containing the parameters for the transformation.

        Returns
        -------
        Any
            The transformed output data.

        Note
        ----
        This method is expected to be implemented by a subclass of the
        ``Transform`` class. It raises a ``NotImplementedError`` to indicate
        that the subclass must provide its own implementation of the
        ``_transform`` method.
        """
        raise NotImplementedError

    def forward(self, *inputs: Any) -> Any:
        """
        Parameters
        ----------
        inputs : tuple
            The input tensors to be transformed. Can have multiple tensors.

        Returns
        -------
        tensor
            The transformed input tensors.
        """
        if len(inputs) > 1:
            flattened, spec = torch.utils._pytree.tree_flatten(inputs)
        else:
            flattened, spec = torch.utils._pytree.tree_flatten(inputs[0])

        self._check_inputs(flattened)

        transformables = self._transformables(flattened)

        inputs = []

        for x, transformable in zip(flattened, transformables, strict=False):
            if transformable:
                inputs = [*inputs, x]

        inputs = self._get_params(inputs)

        ys = []

        for x, transformable in zip(flattened, transformables, strict=False):
            if transformable:
                y = self._transform(x, inputs)
            else:
                y = x

            ys = [*ys, y]

        return torch.utils._pytree.tree_unflatten(ys, spec)

    def _transformables(self, inputs: List[Any]) -> List[bool]:
        """
        Parameters
        ----------
        inputs : List[Any]
            List of input objects to be checked for transformation.

        Returns
        -------
        transformables : List[bool]
            List indicating whether each input object is transformable or not.
            ``True`` if an object can be transformed, ``False`` otherwise.
        """
        # NOTE:
        #   Heuristic for transforming anonymous tensor inputs:
        #
        #       1.  Anonymous tensors, i.e., non-:class:`Feature` tensors,
        #           are passed through if there is an explicit feature in the
        #           sample.
        #
        #       2.  If there is no explicit feature the sample, only the first
        #           encountered anonymous tensor is transformed, while the rest
        #           are passed. The order is defined by the returned
        #           `flat_inputs` of `tree_flatten`, which recurses
        #           depth-first through the input.
        #
        #   The heuristic should work well for most people in practice. The
        #   only case where it doesn't is if someone tries to transform
        #   multiple anonymous features at the same time, expecting them all
        #   to be treated as named features.
        transformables = []

        transform_anonymous_feature = False

        for input in inputs:
            for t in [str]:
                if isinstance(input, t) if isinstance(t, type) else t(input):
                    transform_anonymous_feature = True

        for input in inputs:
            transformable = True

            checked = False

            for t in self._transformed_types:
                if isinstance(input, t) if isinstance(t, type) else t(input):
                    checked = True

            if not checked:
                transformable = False
            elif isinstance(input, Tensor) and not isinstance(input, Feature):
                if transform_anonymous_feature:
                    transform_anonymous_feature = False
                else:
                    transformable = False

            transformables = [*transformables, transformable]

        return transformables

    def extra_repr(self) -> str:
        """
        Returns
        -------
        str
            A string representation of the extra configuration attributes of
            the Transform module.
        """
        extra = []

        for name, value in self.__dict__.items():
            if name.startswith("_") or name == "training":
                continue

            if not isinstance(
                value,
                (Enum, bool, float, int, list, str, tuple),
            ):
                continue

            extra.append(f"{name}={value}")

        return ", ".join(extra)



import beignet
import torch


def test_laguerre_polynomial_roots():
    torch.testing.assert_close(
        beignet.laguerre_polynomial_roots(
            torch.tensor([1.0]),
        ),
        torch.tensor([]),
    )

    torch.testing.assert_close(
        beignet.laguerre_polynomial_roots(
            torch.tensor([0.0, 1.0]),
        ),
        torch.tensor([1.0]),
    )

    for index in range(2, 5):
        torch.testing.assert_close(
            beignet.trim_laguerre_polynomial_coefficients(
                beignet.laguerre_polynomial_roots(
                    beignet.laguerre_polynomial_from_roots(
                        torch.linspace(0, 3, index),
                    ),
                ),
                tol=0.000001,
            ),
            beignet.trim_laguerre_polynomial_coefficients(
                torch.linspace(0, 3, index),
                tol=0.000001,
            ),
        )



import math

import beignet
import torch


def test_polynomial_from_roots():
    coefficients = [
        torch.tensor([1.0]),
        torch.tensor([0.0, 1]),
        torch.tensor([-1.0, 0, 2]),
        torch.tensor([0.0, -3, 0, 4]),
        torch.tensor([1.0, 0, -8, 0, 8]),
        torch.tensor([0.0, 5, 0, -20, 0, 16]),
        torch.tensor([-1.0, 0, 18, 0, -48, 0, 32]),
        torch.tensor([0.0, -7, 0, 56, 0, -112, 0, 64]),
        torch.tensor([1.0, 0, -32, 0, 160, 0, -256, 0, 128]),
        torch.tensor([0.0, 9, 0, -120, 0, 432, 0, -576, 0, 256]),
    ]

    torch.testing.assert_close(
        beignet.trim_polynomial_coefficients(
            beignet.polynomial_from_roots(
                torch.tensor([]),
            ),
            tol=0.000001,
        ),
        torch.tensor([1.0]),
    )

    for index in range(1, 5):
        input = torch.linspace(-math.pi, 0.0, 2 * index + 1)

        input = input[1::2]

        input = torch.cos(input)

        output = beignet.polynomial_from_roots(input) * 2 ** (index - 1)

        torch.testing.assert_close(
            beignet.trim_polynomial_coefficients(
                output,
                tol=0.000001,
            ),
            beignet.trim_polynomial_coefficients(
                coefficients[index],
                tol=0.000001,
            ),
        )



import math

import beignet.special
import hypothesis
import hypothesis.strategies
import scipy
import torch


@hypothesis.strategies.composite
def _strategy(function):
    dtype = function(hypothesis.strategies.sampled_from([torch.float64, torch.float32]))

    # avoid overflow of exp(y^2)
    limit = math.sqrt(math.log(torch.finfo(dtype).max)) - 0.2

    x = function(
        hypothesis.strategies.floats(
            min_value=-limit, max_value=limit, allow_nan=False, allow_infinity=False
        )
    )
    y = function(
        hypothesis.strategies.floats(
            min_value=-limit, max_value=limit, allow_nan=False, allow_infinity=False
        )
    )

    input = torch.complex(torch.tensor(x, dtype=dtype), torch.tensor(y, dtype=dtype))

    return input, scipy.special.wofz(input)


@hypothesis.given(_strategy())
def test_faddeeva_w(data):
    input, output = data

    if input.dtype == torch.complex64:
        rtol, atol = 1e-5, 1e-5
    elif input.dtype == torch.complex128:
        rtol, atol = 1e-10, 1e-10
    else:
        rtol, atol = None, None

    torch.testing.assert_close(
        beignet.special.faddeeva_w(input), output, rtol=rtol, atol=atol
    )


def test_faddeeva_w_propagates_nan():
    input = torch.complex(torch.tensor(torch.nan), torch.tensor(torch.nan))
    output = beignet.special.faddeeva_w(input)
    torch.testing.assert_close(input, output, equal_nan=True)



from typing import Literal

import torch
from torch import Tensor


def multiply_polynomial_by_x(
    input: Tensor,
    mode: Literal["full", "same", "valid"] = "full",
) -> Tensor:
    r"""
    Parameters
    ----------
    input : Tensor
        Polynomial coefficients.

    mode : Literal["full", "same", "valid"]

    Returns
    -------
    output : Tensor
        Polynomial coefficients of the product of the polynomial and the
        independent variable.
    """
    input = torch.atleast_1d(input)

    output = torch.zeros(input.shape[0] + 1, dtype=input.dtype)

    output[1:] = input

    if mode == "same":
        output = output[: input.shape[0]]

    return output



# import beignet
# import torch
#
#
# def test_gauss_legendre_quadrature():
#     x, w = beignet.gauss_legendre_quadrature(100)
#
#     v = beignet.legvander(
#         x,
#         degree=torch.tensor([99]),
#     )
#
#     vv = (v.T * w) @ v
#
#     vd = 1 / torch.sqrt(vv.diagonal())
#     vv = vd[:, None] * vv * vd
#
#     torch.testing.assert_close(
#         vv,
#         torch.eye(100),
#     )
#
#     torch.testing.assert_close(w.sum(), 2.0)



import importlib.metadata
from importlib.metadata import PackageNotFoundError

try:
    __version__ = importlib.metadata.version("beignet")
except PackageNotFoundError:
    __version__ = None

from ._add_chebyshev_polynomial import add_chebyshev_polynomial
from ._add_laguerre_polynomial import add_laguerre_polynomial
from ._add_legendre_polynomial import add_legendre_polynomial
from ._add_physicists_hermite_polynomial import add_physicists_hermite_polynomial
from ._add_polynomial import add_polynomial
from ._add_probabilists_hermite_polynomial import add_probabilists_hermite_polynomial
from ._apply_euler_angle import apply_euler_angle
from ._apply_quaternion import (
    apply_quaternion,
)
from ._apply_rotation_matrix import apply_rotation_matrix
from ._apply_rotation_vector import apply_rotation_vector
from ._apply_transform import apply_transform
from ._chebyshev_extrema import chebyshev_extrema
from ._chebyshev_gauss_quadrature import chebyshev_gauss_quadrature
from ._chebyshev_interpolation import chebyshev_interpolation
from ._chebyshev_polynomial_companion import chebyshev_polynomial_companion
from ._chebyshev_polynomial_domain import chebyshev_polynomial_domain
from ._chebyshev_polynomial_from_roots import chebyshev_polynomial_from_roots
from ._chebyshev_polynomial_one import chebyshev_polynomial_one
from ._chebyshev_polynomial_power import chebyshev_polynomial_power
from ._chebyshev_polynomial_roots import chebyshev_polynomial_roots
from ._chebyshev_polynomial_to_polynomial import chebyshev_polynomial_to_polynomial
from ._chebyshev_polynomial_vandermonde import chebyshev_polynomial_vandermonde
from ._chebyshev_polynomial_vandermonde_2d import chebyshev_polynomial_vandermonde_2d
from ._chebyshev_polynomial_vandermonde_3d import chebyshev_polynomial_vandermonde_3d
from ._chebyshev_polynomial_weight import chebyshev_polynomial_weight
from ._chebyshev_polynomial_x import chebyshev_polynomial_x
from ._chebyshev_polynomial_zero import chebyshev_polynomial_zero
from ._chebyshev_zeros import chebyshev_zeros
from ._compose_euler_angle import compose_euler_angle
from ._compose_quaternion import compose_quaternion
from ._compose_rotation_matrix import compose_rotation_matrix
from ._compose_rotation_vector import compose_rotation_vector
from ._differentiate_chebyshev_polynomial import differentiate_chebyshev_polynomial
from ._differentiate_laguerre_polynomial import differentiate_laguerre_polynomial
from ._differentiate_legendre_polynomial import differentiate_legendre_polynomial
from ._differentiate_physicists_hermite_polynomial import (
    differentiate_physicists_hermite_polynomial,
)
from ._differentiate_polynomial import differentiate_polynomial
from ._differentiate_probabilists_hermite_polynomial import (
    differentiate_probabilists_hermite_polynomial,
)
from ._divide_chebyshev_polynomial import divide_chebyshev_polynomial
from ._divide_laguerre_polynomial import divide_laguerre_polynomial
from ._divide_legendre_polynomial import divide_legendre_polynomial
from ._divide_physicists_hermite_polynomial import divide_physicists_hermite_polynomial
from ._divide_polynomial import divide_polynomial
from ._divide_probabilists_hermite_polynomial import (
    divide_probabilists_hermite_polynomial,
)
from ._euler_angle_identity import euler_angle_identity
from ._euler_angle_magnitude import euler_angle_magnitude
from ._euler_angle_mean import euler_angle_mean
from ._euler_angle_to_quaternion import (
    euler_angle_to_quaternion,
)
from ._euler_angle_to_rotation_matrix import euler_angle_to_rotation_matrix
from ._euler_angle_to_rotation_vector import euler_angle_to_rotation_vector
from ._evaluate_chebyshev_polynomial import evaluate_chebyshev_polynomial
from ._evaluate_chebyshev_polynomial_2d import evaluate_chebyshev_polynomial_2d
from ._evaluate_chebyshev_polynomial_3d import evaluate_chebyshev_polynomial_3d
from ._evaluate_chebyshev_polynomial_cartesian_2d import (
    evaluate_chebyshev_polynomial_cartesian_2d,
)
from ._evaluate_chebyshev_polynomial_cartesian_3d import (
    evaluate_chebyshev_polynomial_cartesian_3d,
)
from ._evaluate_laguerre_polynomial import evaluate_laguerre_polynomial
from ._evaluate_laguerre_polynomial_2d import evaluate_laguerre_polynomial_2d
from ._evaluate_laguerre_polynomial_3d import evaluate_laguerre_polynomial_3d
from ._evaluate_laguerre_polynomial_cartesian_2d import (
    evaluate_laguerre_polynomial_cartesian_2d,
)
from ._evaluate_laguerre_polynomial_cartesian_3d import (
    evaluate_laguerre_polynomial_cartesian_3d,
)
from ._evaluate_legendre_polynomial import evaluate_legendre_polynomial
from ._evaluate_legendre_polynomial_2d import evaluate_legendre_polynomial_2d
from ._evaluate_legendre_polynomial_3d import evaluate_legendre_polynomial_3d
from ._evaluate_legendre_polynomial_cartesian_2d import (
    evaluate_legendre_polynomial_cartesian_2d,
)
from ._evaluate_legendre_polynomial_cartesian_3d import (
    evaluate_legendre_polynomial_cartesian_3d,
)
from ._evaluate_physicists_hermite_polynomial import (
    evaluate_physicists_hermite_polynomial,
)
from ._evaluate_physicists_hermite_polynomial_2d import (
    evaluate_physicists_hermite_polynomial_2d,
)
from ._evaluate_physicists_hermite_polynomial_3d import (
    evaluate_physicists_hermite_polynomial_3d,
)
from ._evaluate_physicists_hermite_polynomial_cartesian_2d import (
    evaluate_physicists_hermite_polynomial_cartesian_2d,
)
from ._evaluate_physicists_hermite_polynomial_cartesian_3d import (
    evaluate_physicists_hermite_polynomial_cartesian_3d,
)
from ._evaluate_polynomial import evaluate_polynomial
from ._evaluate_polynomial_2d import evaluate_polynomial_2d
from ._evaluate_polynomial_3d import evaluate_polynomial_3d
from ._evaluate_polynomial_cartesian_2d import evaluate_polynomial_cartesian_2d
from ._evaluate_polynomial_cartesian_3d import evaluate_polynomial_cartesian_3d
from ._evaluate_polynomial_from_roots import evaluate_polynomial_from_roots
from ._evaluate_probabilists_hermite_polynomial import (
    evaluate_probabilists_hermite_polynomial,
)
from ._evaluate_probabilists_hermite_polynomial_2d import (
    evaluate_probabilists_hermite_polynomial_2d,
)
from ._evaluate_probabilists_hermite_polynomial_3d import (
    evaluate_probabilists_hermite_polynomial_3d,
)
from ._evaluate_probabilists_hermite_polynomial_cartersian_2d import (
    evaluate_probabilists_hermite_polynomial_cartersian_2d,
)
from ._evaluate_probabilists_hermite_polynomial_cartersian_3d import (
    evaluate_probabilists_hermite_polynomial_cartersian_3d,
)
from ._farthest_first_traversal import farthest_first_traversal
from ._fit_chebyshev_polynomial import fit_chebyshev_polynomial
from ._fit_laguerre_polynomial import fit_laguerre_polynomial
from ._fit_legendre_polynomial import fit_legendre_polynomial
from ._fit_physicists_hermite_polynomial import fit_physicists_hermite_polynomial
from ._fit_polynomial import fit_polynomial
from ._fit_probabilists_hermite_polynomial import fit_probabilists_hermite_polynomial
from ._gauss_laguerre_quadrature import gauss_laguerre_quadrature
from ._gauss_legendre_quadrature import gauss_legendre_quadrature
from ._gauss_physicists_hermite_polynomial_quadrature import (
    gauss_physicists_hermite_polynomial_quadrature,
)
from ._gauss_probabilists_hermite_polynomial_quadrature import (
    gauss_probabilists_hermite_polynomial_quadrature,
)
from ._integrate_chebyshev_polynomial import integrate_chebyshev_polynomial
from ._integrate_laguerre_polynomial import integrate_laguerre_polynomial
from ._integrate_legendre_polynomial import integrate_legendre_polynomial
from ._integrate_physicists_hermite_polynomial import (
    integrate_physicists_hermite_polynomial,
)
from ._integrate_polynomial import integrate_polynomial
from ._integrate_probabilists_hermite_polynomial import (
    integrate_probabilists_hermite_polynomial,
)
from ._invert_euler_angle import invert_euler_angle
from ._invert_quaternion import invert_quaternion
from ._invert_rotation_matrix import invert_rotation_matrix
from ._invert_rotation_vector import invert_rotation_vector
from ._invert_transform import invert_transform
from ._laguerre_polynomial_companion import laguerre_polynomial_companion
from ._laguerre_polynomial_domain import laguerre_polynomial_domain
from ._laguerre_polynomial_from_roots import laguerre_polynomial_from_roots
from ._laguerre_polynomial_one import laguerre_polynomial_one
from ._laguerre_polynomial_power import laguerre_polynomial_power
from ._laguerre_polynomial_roots import laguerre_polynomial_roots
from ._laguerre_polynomial_to_polynomial import laguerre_polynomial_to_polynomial
from ._laguerre_polynomial_vandermonde import laguerre_polynomial_vandermonde
from ._laguerre_polynomial_vandermonde_2d import laguerre_polynomial_vandermonde_2d
from ._laguerre_polynomial_vandermonde_3d import laguerre_polynomial_vandermonde_3d
from ._laguerre_polynomial_weight import laguerre_polynomial_weight
from ._laguerre_polynomial_x import laguerre_polynomial_x
from ._laguerre_polynomial_zero import laguerre_polynomial_zero
from ._legendre_polynomial_companion import legendre_polynomial_companion
from ._legendre_polynomial_domain import legendre_polynomial_domain
from ._legendre_polynomial_from_roots import legendre_polynomial_from_roots
from ._legendre_polynomial_one import legendre_polynomial_one
from ._legendre_polynomial_power import legendre_polynomial_power
from ._legendre_polynomial_roots import legendre_polynomial_roots
from ._legendre_polynomial_to_polynomial import legendre_polynomial_to_polynomial
from ._legendre_polynomial_vandermonde import legendre_polynomial_vandermonde
from ._legendre_polynomial_vandermonde_2d import legendre_polynomial_vandermonde_2d
from ._legendre_polynomial_vandermonde_3d import legendre_polynomial_vandermonde_3d
from ._legendre_polynomial_weight import legendre_polynomial_weight
from ._legendre_polynomial_x import legendre_polynomial_x
from ._legendre_polynomial_zero import legendre_polynomial_zero
from ._lennard_jones_potential import lennard_jones_potential
from ._linear_chebyshev_polynomial import linear_chebyshev_polynomial
from ._linear_laguerre_polynomial import linear_laguerre_polynomial
from ._linear_legendre_polynomial import linear_legendre_polynomial
from ._linear_physicists_hermite_polynomial import linear_physicists_hermite_polynomial
from ._linear_polynomial import linear_polynomial
from ._linear_probabilists_hermite_polynomial import (
    linear_probabilists_hermite_polynomial,
)
from ._multiply_chebyshev_polynomial import multiply_chebyshev_polynomial
from ._multiply_chebyshev_polynomial_by_x import multiply_chebyshev_polynomial_by_x
from ._multiply_laguerre_polynomial import multiply_laguerre_polynomial
from ._multiply_laguerre_polynomial_by_x import multiply_laguerre_polynomial_by_x
from ._multiply_legendre_polynomial import multiply_legendre_polynomial
from ._multiply_legendre_polynomial_by_x import multiply_legendre_polynomial_by_x
from ._multiply_physicists_hermite_polynomial import (
    multiply_physicists_hermite_polynomial,
)
from ._multiply_physicists_hermite_polynomial_by_x import (
    multiply_physicists_hermite_polynomial_by_x,
)
from ._multiply_polynomial import multiply_polynomial
from ._multiply_polynomial_by_x import multiply_polynomial_by_x
from ._multiply_probabilists_hermite_polynomial import (
    multiply_probabilists_hermite_polynomial,
)
from ._multiply_probabilists_hermite_polynomial_by_x import (
    multiply_probabilists_hermite_polynomial_by_x,
)
from ._physicists_hermite_polynomial_companion import (
    physicists_hermite_polynomial_companion,
)
from ._physicists_hermite_polynomial_domain import physicists_hermite_polynomial_domain
from ._physicists_hermite_polynomial_from_roots import (
    physicists_hermite_polynomial_from_roots,
)
from ._physicists_hermite_polynomial_one import physicists_hermite_polynomial_one
from ._physicists_hermite_polynomial_power import physicists_hermite_polynomial_power
from ._physicists_hermite_polynomial_roots import physicists_hermite_polynomial_roots
from ._physicists_hermite_polynomial_to_polynomial import (
    physicists_hermite_polynomial_to_polynomial,
)
from ._physicists_hermite_polynomial_vandermonde import (
    physicists_hermite_polynomial_vandermonde,
)
from ._physicists_hermite_polynomial_vandermonde_2d import (
    physicists_hermite_polynomial_vandermonde_2d,
)
from ._physicists_hermite_polynomial_vandermonde_3d import (
    physicists_hermite_polynomial_vandermonde_3d,
)
from ._physicists_hermite_polynomial_weight import physicists_hermite_polynomial_weight
from ._physicists_hermite_polynomial_x import physicists_hermite_polynomial_x
from ._physicists_hermite_polynomial_zero import physicists_hermite_polynomial_zero
from ._polynomial_companion import polynomial_companion
from ._polynomial_domain import polynomial_domain
from ._polynomial_from_roots import polynomial_from_roots
from ._polynomial_one import polynomial_one
from ._polynomial_power import polynomial_power
from ._polynomial_roots import polynomial_roots
from ._polynomial_to_chebyshev_polynomial import polynomial_to_chebyshev_polynomial
from ._polynomial_to_laguerre_polynomial import polynomial_to_laguerre_polynomial
from ._polynomial_to_legendre_polynomial import polynomial_to_legendre_polynomial
from ._polynomial_to_physicists_hermite_polynomial import (
    polynomial_to_physicists_hermite_polynomial,
)
from ._polynomial_to_probabilists_hermite_polynomial import (
    polynomial_to_probabilists_hermite_polynomial,
)
from ._polynomial_vandermonde import polynomial_vandermonde
from ._polynomial_vandermonde_2d import polynomial_vandermonde_2d
from ._polynomial_vandermonde_3d import polynomial_vandermonde_3d
from ._polynomial_x import polynomial_x
from ._polynomial_zero import polynomial_zero
from ._probabilists_hermite_polynomial_companion import (
    probabilists_hermite_polynomial_companion,
)
from ._probabilists_hermite_polynomial_domain import (
    probabilists_hermite_polynomial_domain,
)
from ._probabilists_hermite_polynomial_from_roots import (
    probabilists_hermite_polynomial_from_roots,
)
from ._probabilists_hermite_polynomial_one import probabilists_hermite_polynomial_one
from ._probabilists_hermite_polynomial_power import (
    probabilists_hermite_polynomial_power,
)
from ._probabilists_hermite_polynomial_roots import (
    probabilists_hermite_polynomial_roots,
)
from ._probabilists_hermite_polynomial_to_polynomial import (
    probabilists_hermite_polynomial_to_polynomial,
)
from ._probabilists_hermite_polynomial_vandermonde import (
    probabilists_hermite_polynomial_vandermonde,
)
from ._probabilists_hermite_polynomial_vandermonde_2d import (
    probabilists_hermite_polynomial_vandermonde_2d,
)
from ._probabilists_hermite_polynomial_vandermonde_3d import (
    probabilists_hermite_polynomial_vandermonde_3d,
)
from ._probabilists_hermite_polynomial_weight import (
    probabilists_hermite_polynomial_weight,
)
from ._probabilists_hermite_polynomial_x import probabilists_hermite_polynomial_x
from ._probabilists_hermite_polynomial_zero import probabilists_hermite_polynomial_zero
from ._quaternion_identity import quaternion_identity
from ._quaternion_magnitude import quaternion_magnitude
from ._quaternion_mean import quaternion_mean
from ._quaternion_slerp import quaternion_slerp
from ._quaternion_to_euler_angle import (
    quaternion_to_euler_angle,
)
from ._quaternion_to_rotation_matrix import (
    quaternion_to_rotation_matrix,
)
from ._quaternion_to_rotation_vector import (
    quaternion_to_rotation_vector,
)
from ._random_euler_angle import random_euler_angle
from ._random_quaternion import random_quaternion
from ._random_rotation_matrix import random_rotation_matrix
from ._random_rotation_vector import random_rotation_vector
from ._rotation_matrix_identity import rotation_matrix_identity
from ._rotation_matrix_magnitude import rotation_matrix_magnitude
from ._rotation_matrix_mean import rotation_matrix_mean
from ._rotation_matrix_to_euler_angle import rotation_matrix_to_euler_angle
from ._rotation_matrix_to_quaternion import (
    rotation_matrix_to_quaternion,
)
from ._rotation_matrix_to_rotation_vector import (
    rotation_matrix_to_rotation_vector,
)
from ._rotation_vector_identity import rotation_vector_identity
from ._rotation_vector_magnitude import rotation_vector_magnitude
from ._rotation_vector_mean import rotation_vector_mean
from ._rotation_vector_to_euler_angle import rotation_vector_to_euler_angle
from ._rotation_vector_to_quaternion import (
    rotation_vector_to_quaternion,
)
from ._rotation_vector_to_rotation_matrix import (
    rotation_vector_to_rotation_matrix,
)
from ._subtract_chebyshev_polynomial import subtract_chebyshev_polynomial
from ._subtract_laguerre_polynomial import subtract_laguerre_polynomial
from ._subtract_legendre_polynomial import subtract_legendre_polynomial
from ._subtract_physicists_hermite_polynomial import (
    subtract_physicists_hermite_polynomial,
)
from ._subtract_polynomial import subtract_polynomial
from ._subtract_probabilists_hermite_polynomial import (
    subtract_probabilists_hermite_polynomial,
)
from ._translation_identity import translation_identity
from ._trim_chebyshev_polynomial_coefficients import (
    trim_chebyshev_polynomial_coefficients,
)
from ._trim_laguerre_polynomial_coefficients import (
    trim_laguerre_polynomial_coefficients,
)
from ._trim_legendre_polynomial_coefficients import (
    trim_legendre_polynomial_coefficients,
)
from ._trim_physicists_hermite_polynomial_coefficients import (
    trim_physicists_hermite_polynomial_coefficients,
)
from ._trim_polynomial_coefficients import trim_polynomial_coefficients
from ._trim_probabilists_hermite_polynomial_coefficients import (
    trim_probabilists_hermite_polynomial_coefficients,
)
from .special import error_erf, error_erfc

__all__ = [
    "add_chebyshev_polynomial",
    "add_laguerre_polynomial",
    "add_legendre_polynomial",
    "add_physicists_hermite_polynomial",
    "add_polynomial",
    "add_probabilists_hermite_polynomial",
    "apply_euler_angle",
    "apply_quaternion",
    "apply_rotation_matrix",
    "apply_rotation_vector",
    "apply_transform",
    "evaluate_chebyshev_polynomial_cartesian_2d",
    "evaluate_chebyshev_polynomial_cartesian_3d",
    "chebyshev_interpolation",
    "linear_chebyshev_polynomial",
    "multiply_chebyshev_polynomial_by_x",
    "chebyshev_zeros",
    "chebyshev_extrema",
    "chebyshev_gauss_quadrature",
    "chebyshev_polynomial_companion",
    "chebyshev_polynomial_domain",
    "chebyshev_polynomial_from_roots",
    "chebyshev_polynomial_one",
    "chebyshev_polynomial_power",
    "chebyshev_polynomial_roots",
    "chebyshev_polynomial_to_polynomial",
    "chebyshev_polynomial_vandermonde",
    "chebyshev_polynomial_vandermonde_2d",
    "chebyshev_polynomial_vandermonde_3d",
    "chebyshev_polynomial_weight",
    "chebyshev_polynomial_x",
    "chebyshev_polynomial_zero",
    "compose_euler_angle",
    "compose_quaternion",
    "compose_rotation_matrix",
    "compose_rotation_vector",
    "differentiate_chebyshev_polynomial",
    "differentiate_laguerre_polynomial",
    "differentiate_legendre_polynomial",
    "differentiate_physicists_hermite_polynomial",
    "differentiate_polynomial",
    "differentiate_probabilists_hermite_polynomial",
    "divide_chebyshev_polynomial",
    "divide_laguerre_polynomial",
    "divide_legendre_polynomial",
    "divide_physicists_hermite_polynomial",
    "divide_polynomial",
    "divide_probabilists_hermite_polynomial",
    "euler_angle_identity",
    "euler_angle_magnitude",
    "euler_angle_mean",
    "euler_angle_to_quaternion",
    "euler_angle_to_rotation_matrix",
    "euler_angle_to_rotation_vector",
    "evaluate_chebyshev_polynomial",
    "evaluate_chebyshev_polynomial_2d",
    "evaluate_chebyshev_polynomial_3d",
    "evaluate_laguerre_polynomial",
    "evaluate_laguerre_polynomial_2d",
    "evaluate_laguerre_polynomial_3d",
    "evaluate_legendre_polynomial",
    "evaluate_legendre_polynomial_2d",
    "evaluate_legendre_polynomial_3d",
    "evaluate_physicists_hermite_polynomial",
    "evaluate_physicists_hermite_polynomial_2d",
    "evaluate_physicists_hermite_polynomial_3d",
    "evaluate_polynomial",
    "evaluate_polynomial_2d",
    "evaluate_polynomial_3d",
    "evaluate_polynomial_from_roots",
    "evaluate_probabilists_hermite_polynomial",
    "evaluate_probabilists_hermite_polynomial_2d",
    "evaluate_probabilists_hermite_polynomial_3d",
    "farthest_first_traversal",
    "fit_chebyshev_polynomial",
    "fit_laguerre_polynomial",
    "fit_legendre_polynomial",
    "fit_physicists_hermite_polynomial",
    "fit_polynomial",
    "fit_probabilists_hermite_polynomial",
    "gauss_laguerre_quadrature",
    "gauss_legendre_quadrature",
    "gauss_physicists_hermite_polynomial_quadrature",
    "gauss_probabilists_hermite_polynomial_quadrature",
    "evaluate_probabilists_hermite_polynomial_cartersian_2d",
    "evaluate_probabilists_hermite_polynomial_cartersian_3d",
    "linear_probabilists_hermite_polynomial",
    "multiply_probabilists_hermite_polynomial_by_x",
    "evaluate_physicists_hermite_polynomial_cartesian_2d",
    "evaluate_physicists_hermite_polynomial_cartesian_3d",
    "linear_physicists_hermite_polynomial",
    "multiply_physicists_hermite_polynomial_by_x",
    "integrate_chebyshev_polynomial",
    "integrate_laguerre_polynomial",
    "integrate_legendre_polynomial",
    "integrate_physicists_hermite_polynomial",
    "integrate_polynomial",
    "integrate_probabilists_hermite_polynomial",
    "invert_euler_angle",
    "invert_quaternion",
    "invert_rotation_matrix",
    "invert_rotation_vector",
    "invert_transform",
    "evaluate_laguerre_polynomial_cartesian_2d",
    "evaluate_laguerre_polynomial_cartesian_3d",
    "linear_laguerre_polynomial",
    "multiply_laguerre_polynomial_by_x",
    "laguerre_polynomial_companion",
    "laguerre_polynomial_domain",
    "laguerre_polynomial_from_roots",
    "laguerre_polynomial_one",
    "laguerre_polynomial_power",
    "laguerre_polynomial_roots",
    "laguerre_polynomial_to_polynomial",
    "laguerre_polynomial_vandermonde",
    "laguerre_polynomial_vandermonde_2d",
    "laguerre_polynomial_vandermonde_3d",
    "laguerre_polynomial_weight",
    "laguerre_polynomial_x",
    "laguerre_polynomial_zero",
    "legendre_polynomial_companion",
    "legendre_polynomial_domain",
    "legendre_polynomial_from_roots",
    "legendre_polynomial_one",
    "legendre_polynomial_power",
    "legendre_polynomial_roots",
    "legendre_polynomial_to_polynomial",
    "legendre_polynomial_vandermonde",
    "legendre_polynomial_vandermonde_2d",
    "legendre_polynomial_vandermonde_3d",
    "legendre_polynomial_weight",
    "legendre_polynomial_x",
    "legendre_polynomial_zero",
    "evaluate_legendre_polynomial_cartesian_2d",
    "evaluate_legendre_polynomial_cartesian_3d",
    "linear_legendre_polynomial",
    "multiply_legendre_polynomial_by_x",
    "lennard_jones_potential",
    "multiply_chebyshev_polynomial",
    "multiply_laguerre_polynomial",
    "multiply_legendre_polynomial",
    "multiply_physicists_hermite_polynomial",
    "multiply_polynomial",
    "multiply_probabilists_hermite_polynomial",
    "physicists_hermite_polynomial_companion",
    "physicists_hermite_polynomial_domain",
    "physicists_hermite_polynomial_from_roots",
    "physicists_hermite_polynomial_one",
    "physicists_hermite_polynomial_power",
    "physicists_hermite_polynomial_roots",
    "physicists_hermite_polynomial_to_polynomial",
    "physicists_hermite_polynomial_vandermonde",
    "physicists_hermite_polynomial_vandermonde_2d",
    "physicists_hermite_polynomial_vandermonde_3d",
    "physicists_hermite_polynomial_weight",
    "physicists_hermite_polynomial_x",
    "physicists_hermite_polynomial_zero",
    "evaluate_polynomial_cartesian_2d",
    "evaluate_polynomial_cartesian_3d",
    "linear_polynomial",
    "multiply_polynomial_by_x",
    "polynomial_companion",
    "polynomial_domain",
    "polynomial_from_roots",
    "polynomial_one",
    "polynomial_power",
    "polynomial_roots",
    "polynomial_to_chebyshev_polynomial",
    "polynomial_to_laguerre_polynomial",
    "polynomial_to_legendre_polynomial",
    "polynomial_to_physicists_hermite_polynomial",
    "polynomial_to_probabilists_hermite_polynomial",
    "polynomial_vandermonde",
    "polynomial_vandermonde_2d",
    "polynomial_vandermonde_3d",
    "polynomial_x",
    "polynomial_zero",
    "probabilists_hermite_polynomial_companion",
    "probabilists_hermite_polynomial_domain",
    "probabilists_hermite_polynomial_from_roots",
    "probabilists_hermite_polynomial_one",
    "probabilists_hermite_polynomial_power",
    "probabilists_hermite_polynomial_roots",
    "probabilists_hermite_polynomial_to_polynomial",
    "probabilists_hermite_polynomial_vandermonde",
    "probabilists_hermite_polynomial_vandermonde_2d",
    "probabilists_hermite_polynomial_vandermonde_3d",
    "probabilists_hermite_polynomial_weight",
    "probabilists_hermite_polynomial_x",
    "probabilists_hermite_polynomial_zero",
    "quaternion_identity",
    "quaternion_magnitude",
    "quaternion_mean",
    "quaternion_slerp",
    "quaternion_to_euler_angle",
    "quaternion_to_rotation_matrix",
    "quaternion_to_rotation_vector",
    "random_euler_angle",
    "random_quaternion",
    "random_rotation_matrix",
    "random_rotation_vector",
    "rotation_matrix_identity",
    "rotation_matrix_magnitude",
    "rotation_matrix_mean",
    "rotation_matrix_to_euler_angle",
    "rotation_matrix_to_quaternion",
    "rotation_matrix_to_rotation_vector",
    "rotation_vector_identity",
    "rotation_vector_magnitude",
    "rotation_vector_mean",
    "rotation_vector_to_euler_angle",
    "rotation_vector_to_quaternion",
    "rotation_vector_to_rotation_matrix",
    "subtract_chebyshev_polynomial",
    "subtract_laguerre_polynomial",
    "subtract_legendre_polynomial",
    "subtract_physicists_hermite_polynomial",
    "subtract_polynomial",
    "subtract_probabilists_hermite_polynomial",
    "translation_identity",
    "trim_chebyshev_polynomial_coefficients",
    "trim_laguerre_polynomial_coefficients",
    "trim_legendre_polynomial_coefficients",
    "trim_physicists_hermite_polynomial_coefficients",
    "trim_polynomial_coefficients",
    "trim_probabilists_hermite_polynomial_coefficients",
]



from pathlib import Path
from typing import Callable, Tuple, Union

from pandas import DataFrame

from beignet.transforms import Transform

from ._atom3d_dataset import ATOM3DDataset


class ATOM3DPPIDataset(ATOM3DDataset):
    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = False,
        transform: Union[Callable, Transform, None] = None,
        target_transform: Union[Callable, Transform, None] = None,
        joint_transform_fn: Union[Callable, Transform, None] = None,
    ):
        super().__init__(
            root,
            "raw/DIPS/data",
            "https://zenodo.org/record/4911102/files/PPI-raw.tar.gz",
            "PPI",
            checksum="621977d132b39957e3480a24a30a7358",
            download=download,
        )

        self._transform_fn = transform

        self._target_transform_fn = target_transform

        self._joint_transform_fn = joint_transform_fn

    def __getitem__(self, index: int) -> Tuple[DataFrame, DataFrame]:
        item = super().__getitem__(index)

        features = DataFrame(**item["atoms_pairs"])

        target = DataFrame(**item["atoms_neighbors"])

        if self._joint_transform_fn is not None:
            features, target = self._joint_transform_fn(features, target)

        if self._transform_fn is not None:
            features = self._transform_fn(features)

        if self._target_transform_fn is not None:
            target = self._target_transform_fn(target)

        return features, target



import math

import beignet
import torch


def test_probabilists_hermite_polynomial_from_roots():
    torch.testing.assert_close(
        beignet.trim_probabilists_hermite_polynomial_coefficients(
            beignet.probabilists_hermite_polynomial_from_roots(
                torch.tensor([]),
            ),
            tol=0.000001,
        ),
        torch.tensor([1.0]),
    )

    for i in range(1, 5):
        roots = torch.cos(torch.linspace(-math.pi, 0, 2 * i + 1)[1::2])

        pol = beignet.probabilists_hermite_polynomial_from_roots(roots)

        assert len(pol) == i + 1

        torch.testing.assert_close(
            beignet.probabilists_hermite_polynomial_to_polynomial(pol)[-1],
            torch.tensor(1.0),
        )

        # torch.testing.assert_close(
        #     beignet.hermeval(roots, pol),
        #     torch.tensor([0.0]),
        # )



import beignet
import hypothesis.strategies


@hypothesis.strategies.composite
def _strategy(function):
    size = function(
        hypothesis.strategies.integers(
            min_value=1,
            max_value=8,
        ),
    )

    axes = function(
        hypothesis.strategies.sampled_from(
            [
                "xyz",
                "xzy",
                "yxz",
                "yzx",
                "zxy",
                "zyx",
                "XYZ",
                "XZY",
                "YXZ",
                "YZX",
                "ZXY",
                "ZYX",
            ]
        ),
    )

    degrees = function(hypothesis.strategies.booleans())

    return (
        {
            "size": size,
            "axes": axes,
            "degrees": degrees,
        },
        None,
    )


@hypothesis.given(_strategy())
def test_random_euler_angle(data):
    parameters, _ = data

    assert beignet.random_euler_angle(
        **parameters,
    ).shape == (parameters["size"], 3)



import beignet
import torch


def test_physicists_hermite_polynomial_to_polynomial():
    coefficients = [
        torch.tensor([1.0]),
        torch.tensor([0.0, 2]),
        torch.tensor([-2.0, 0, 4]),
        torch.tensor([0.0, -12, 0, 8]),
        torch.tensor([12.0, 0, -48, 0, 16]),
        torch.tensor([0.0, 120, 0, -160, 0, 32]),
        torch.tensor([-120.0, 0, 720, 0, -480, 0, 64]),
        torch.tensor([0.0, -1680, 0, 3360, 0, -1344, 0, 128]),
        torch.tensor([1680.0, 0, -13440, 0, 13440, 0, -3584, 0, 256]),
        torch.tensor([0.0, 30240, 0, -80640, 0, 48384, 0, -9216, 0, 512]),
    ]

    for index in range(10):
        torch.testing.assert_close(
            beignet.physicists_hermite_polynomial_to_polynomial(
                torch.tensor([0.0] * index + [1.0]),
            ),
            coefficients[index],
        )



import math

import beignet
import torch


def test_chebyshev_polynomial_from_roots():
    torch.testing.assert_close(
        beignet.trim_chebyshev_polynomial_coefficients(
            beignet.chebyshev_polynomial_from_roots(
                torch.tensor([]),
            ),
            tol=0.000001,
        ),
        torch.tensor([1.0]),
    )

    for index in range(1, 5):
        input = beignet.chebyshev_polynomial_from_roots(
            torch.cos(torch.linspace(-math.pi, 0.0, 2 * index + 1)[1::2]),
        )

        input = input * 2 ** (index - 1)

        torch.testing.assert_close(
            beignet.trim_chebyshev_polynomial_coefficients(
                input,
                tol=0.000001,
            ),
            beignet.trim_chebyshev_polynomial_coefficients(
                torch.tensor([0.0] * index + [1.0]),
                tol=0.000001,
            ),
        )



import beignet
import torch


def test_probabilists_hermite_polynomial_vandermonde_2d():
    a, b, c = torch.rand(3, 5) * 2 - 1

    coefficients = torch.rand(2, 3)

    output = beignet.probabilists_hermite_polynomial_vandermonde_2d(
        a,
        b,
        degree=torch.tensor([1, 2]),
    )

    torch.testing.assert_close(
        output @ torch.ravel(coefficients),
        beignet.evaluate_probabilists_hermite_polynomial_2d(
            a,
            b,
            coefficients,
        ),
    )

    output = beignet.probabilists_hermite_polynomial_vandermonde_2d(
        a,
        b,
        degree=torch.tensor([1, 2]),
    )

    assert output.shape == (5, 6)



from torch import Tensor

from ._compose_quaternion import compose_quaternion
from ._quaternion_to_rotation_matrix import quaternion_to_rotation_matrix
from ._rotation_matrix_to_quaternion import rotation_matrix_to_quaternion


def compose_rotation_matrix(
    input: Tensor,
    other: Tensor,
) -> Tensor:
    r"""
    Compose rotation matrices.

    Parameters
    ----------
    input : Tensor, shape=(..., 3, 3)
        Rotation matrices.

    other : Tensor, shape=(..., 3, 3)
        Rotation matrices.

    Returns
    -------
    output : Tensor, shape=(..., 3, 3)
        Composed rotation matrices.
    """
    return quaternion_to_rotation_matrix(
        compose_quaternion(
            rotation_matrix_to_quaternion(input),
            rotation_matrix_to_quaternion(other),
        ),
    )



import numpy
import torch

from ._evaluate_laguerre_polynomial import evaluate_laguerre_polynomial


def integrate_laguerre_polynomial(
    input,
    order=1,
    k=None,
    lower_bound=0,
    scale=1,
    axis=0,
):
    if k is None:
        k = []

    input = torch.atleast_1d(input)

    lower_bound, scale = map(torch.tensor, (lower_bound, scale))

    if not numpy.iterable(k):
        k = [k]
    if len(k) > order:
        raise ValueError

    if lower_bound.ndim != 0:
        raise ValueError

    if scale.ndim != 0:
        raise ValueError

    if order == 0:
        return input

    input = torch.moveaxis(input, axis, 0)
    k = torch.tensor(list(k) + [0] * (order - len(k)))
    k = torch.atleast_1d(k)

    for i in range(order):
        n = input.shape[0]
        input *= scale

        tmp = torch.empty((n + 1,) + input.shape[1:], dtype=input.dtype)

        tmp[0] = input[0]
        tmp[1] = -input[0]

        j = torch.arange(1, n)

        tmp[j] += input[j]
        tmp[j + 1] += -input[j]

        tmp_value = torch.tensor(evaluate_laguerre_polynomial(lower_bound, tmp))
        tmp[0] += k[i] - tmp_value

        input = tmp

    input = torch.moveaxis(input, 0, axis)
    return input



import math

import torch
from torch import Tensor


def chebyshev_polynomial_companion(input: Tensor) -> Tensor:
    input = torch.atleast_1d(input)

    if input.shape[0] < 2:
        raise ValueError

    if input.shape[0] == 2:
        return torch.tensor([[-input[0] / input[1]]])

    n = input.shape[0] - 1

    output = torch.zeros(
        [
            n,
            n,
        ],
        dtype=input.dtype,
    )

    scale = torch.ones([n])

    scale[1:] = math.sqrt(0.5)

    shape = output.shape

    output = torch.reshape(output, [-1])

    x = torch.full([n - 1], 1 / 2)

    x[0] = math.sqrt(0.5)

    output[1 :: n + 1] = x
    output[n :: n + 1] = x

    output = torch.reshape(output, shape)

    output[:, -1] = (
        output[:, -1] + -(input[:-1] / input[-1]) * (scale / scale[-1]) * 0.5
    )

    return output



import math

import beignet
import torch


def test_evaluate_chebyshev_polynomial():
    chebcoefficients = [
        torch.tensor([1]),
        torch.tensor([0, 1]),
        torch.tensor([-1, 0, 2]),
        torch.tensor([0, -3, 0, 4]),
        torch.tensor([1, 0, -8, 0, 8]),
        torch.tensor([0, 5, 0, -20, 0, 16]),
        torch.tensor([-1, 0, 18, 0, -48, 0, 32]),
        torch.tensor([0, -7, 0, 56, 0, -112, 0, 64]),
        torch.tensor([1, 0, -32, 0, 160, 0, -256, 0, 128]),
        torch.tensor([0, 9, 0, -120, 0, 432, 0, -576, 0, 256]),
    ]

    output = beignet.evaluate_chebyshev_polynomial(
        torch.tensor([]),
        torch.tensor([1.0]),
    )

    assert math.prod(output.shape) == 0

    ys = []

    for coefficient in chebcoefficients:
        ys = [
            *ys,
            beignet.evaluate_polynomial(
                torch.linspace(-1, 1, 50),
                coefficient,
            ),
        ]

    for index in range(10):
        torch.testing.assert_close(
            beignet.evaluate_chebyshev_polynomial(
                torch.linspace(-1, 1, 50),
                torch.tensor([0.0] * index + [1.0]),
            ),
            torch.tensor(ys[index]),
        )

    for index in range(3):
        shape = (2,) * index

        input = torch.zeros(shape)

        output = beignet.evaluate_chebyshev_polynomial(
            input,
            torch.tensor([1.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_chebyshev_polynomial(
            input,
            torch.tensor([1.0, 0.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_chebyshev_polynomial(
            input,
            torch.tensor([1.0, 0.0, 0.0]),
        )

        assert output.shape == shape



from ._aqsoldb_solubility_dataset import AqSolDBSolubilityDataset
from ._astrazeneca_clearance_dataset import AstraZenecaClearanceDataset
from ._astrazeneca_lipophilicity_dataset import AstraZenecaLipophilicityDataset
from ._astrazeneca_plasma_protein_binding_rate_dataset import (
    AstraZenecaPlasmaProteinBindingRateDataset,
)
from ._atom3d_dataset import ATOM3DDataset
from ._atom3d_msp_dataset import ATOM3DMSPDataset
from ._atom3d_ppi_dataset import ATOM3DPPIDataset
from ._atom3d_psr_dataset import ATOM3DPSRDataset
from ._atom3d_res_dataset import ATOM3DRESDataset
from ._atom3d_rsr_dataset import ATOM3DRSRDataset
from ._atom3d_smp_dataset import ATOM3DSMPDataset
from ._bindingdb_ic50_dataset import BindingDBIC50Dataset
from ._bindingdb_kd_dataset import BindingDBKdDataset
from ._bindingdb_ki_dataset import BindingDBKiDataset
from ._broccatelli_p_glycoprotein_inhibition_dataset import (
    BroccatelliPGlycoproteinInhibitionDataset,
)
from ._carbon_mangels_cytochrome_p450_2c9_substrate_dataset import (
    CarbonMangelsCytochromeP4502C9SubstrateDataset,
)
from ._carbon_mangels_cytochrome_p450_2d6_substrate_dataset import (
    CarbonMangelsCytochromeP4502D6SubstrateDataset,
)
from ._carbon_mangels_cytochrome_p450_3a4_substrate_dataset import (
    CarbonMangelsCytochromeP4503A4SubstrateDataset,
)
from ._chembl_dataset import ChEMBLDataset
from ._clintox_dataset import ClinToxDataset
from ._data_frame_dataset import DataFrameDataset
from ._davis_dataset import DAVISDataset
from ._disgenet_dataset import DisGeNETDataset
from ._drugcomb_dataset import DrugCombDataset
from ._fasta_dataset import FASTADataset
from ._freesolv_dataset import FreeSolvDataset
from ._gdsc1_dataset import GDSC1Dataset
from ._gdsc2_dataset import GDSC2Dataset
from ._hdf5_trajectory_dataset import HDF5TrajectoryDataset
from ._hou_human_intestinal_absorption_dataset import (
    HouHumanIntestinalAbsorptionDataset,
)
from ._huri_dataset import HuRIDataset
from ._jespersen_iedb_dataset import JespersenIEDBDataset
from ._jespersen_pdb_dataset import JespersenPDBDataset
from ._kiba_dataset import KIBADataset
from ._lmdb_dataset import LMDBDataset
from ._lombardo_volume_of_distribution_at_steady_state_dataset import (
    LombardoVolumeOfDistributionAtSteadyStateDataset,
)
from ._ma_bioavailability_dataset import MaBioavailabilityDataset
from ._martins_blood_brain_barrier_dataset import MartinsBloodBrainBarrierDataset
from ._moses_dataset import MOSESDataset
from ._ncats_pampa_permeability_dataset import NCATSPAMPAPermeabilityDataset
from ._obach_half_life_dataset import ObachHalfLifeDataset
from ._parquet_dataset import ParquetDataset
from ._pdb_trajectory_dataset import PDBTrajectoryDataset
from ._pdbbind_dataset import PDBbindDataset
from ._qm7_dataset import QM7Dataset
from ._qm7b_dataset import QM7bDataset
from ._qm8_dataset import QM8Dataset
from ._qm9_dataset import QM9Dataset
from ._random_euler_angle_dataset import RandomEulerAngleDataset
from ._random_quaternion_dataset import RandomQuaternionDataset
from ._random_rotation_dataset import RandomRotationDataset
from ._random_rotation_matrix_dataset import RandomRotationMatrixDataset
from ._random_rotation_vector_dataset import RandomRotationVectorDataset
from ._real_database_dataset import REALDatabaseDataset
from ._sabdab_dataset import SAbDabDataset
from ._sequence_dataset import SequenceDataset
from ._sized_sequence_dataset import SizedSequenceDataset
from ._skempi_dataset import SKEMPIDataset
from ._swissprot_dataset import SwissProtDataset
from ._tdc_dataset import TDCDataset
from ._therapeutic_antibody_profiler_dataset import TherapeuticAntibodyProfilerDataset
from ._tox21_dataset import Tox21Dataset
from ._toxcast_dataset import ToxCastDataset
from ._trajectory_dataset import TrajectoryDataset
from ._trembl_dataset import TrEMBLDataset
from ._uniprot_dataset import UniProtDataset
from ._uniref50_dataset import UniRef50Dataset
from ._uniref90_dataset import UniRef90Dataset
from ._uniref100_dataset import UniRef100Dataset
from ._uspto_dataset import USPTODataset
from ._uspto_reaction_product_dataset import USPTOReactionProductDataset
from ._veith_cytochrome_p450_1a2_inhibition_dataset import (
    VeithCytochromeP4501A2InhibitionDataset,
)
from ._veith_cytochrome_p450_2c9_inhibition_dataset import (
    VeithCytochromeP4502C9InhibitionDataset,
)
from ._veith_cytochrome_p450_2c19_inhibition_dataset import (
    VeithCytochromeP4502C19InhibitionDataset,
)
from ._veith_cytochrome_p450_2d6_inhibition_dataset import (
    VeithCytochromeP4502D6InhibitionDataset,
)
from ._veith_cytochrome_p450_3a4_inhibition_dataset import (
    VeithCytochromeP4503A4InhibitionDataset,
)
from ._wang_effective_permeability_dataset import WangEffectivePermeabilityDataset
from ._zhu_acute_toxicity_ld50_dataset import ZhuAcuteToxicityLD50Dataset
from ._zinc_dataset import ZINCDataset

__all__ = [
    "ATOM3DDataset",
    "ATOM3DMSPDataset",
    "ATOM3DPPIDataset",
    "ATOM3DPSRDataset",
    "ATOM3DRESDataset",
    "ATOM3DRSRDataset",
    "ATOM3DSMPDataset",
    "AqSolDBSolubilityDataset",
    "AstraZenecaClearanceDataset",
    "AstraZenecaLipophilicityDataset",
    "AstraZenecaPlasmaProteinBindingRateDataset",
    "BindingDBIC50Dataset",
    "BindingDBKdDataset",
    "BindingDBKiDataset",
    "BroccatelliPGlycoproteinInhibitionDataset",
    "CarbonMangelsCytochromeP4502C9SubstrateDataset",
    "CarbonMangelsCytochromeP4502D6SubstrateDataset",
    "CarbonMangelsCytochromeP4503A4SubstrateDataset",
    "ChEMBLDataset",
    "ClinToxDataset",
    "DAVISDataset",
    "DataFrameDataset",
    "DisGeNETDataset",
    "DrugCombDataset",
    "FASTADataset",
    "FreeSolvDataset",
    "GDSC1Dataset",
    "GDSC2Dataset",
    "HDF5TrajectoryDataset",
    "HouHumanIntestinalAbsorptionDataset",
    "HuRIDataset",
    "JespersenIEDBDataset",
    "JespersenPDBDataset",
    "KIBADataset",
    "LMDBDataset",
    "LombardoVolumeOfDistributionAtSteadyStateDataset",
    "MOSESDataset",
    "MaBioavailabilityDataset",
    "MartinsBloodBrainBarrierDataset",
    "NCATSPAMPAPermeabilityDataset",
    "ObachHalfLifeDataset",
    "PDBTrajectoryDataset",
    "PDBbindDataset",
    "ParquetDataset",
    "QM7Dataset",
    "QM7bDataset",
    "QM8Dataset",
    "QM9Dataset",
    "REALDatabaseDataset",
    "RandomEulerAngleDataset",
    "RandomQuaternionDataset",
    "RandomRotationDataset",
    "RandomRotationMatrixDataset",
    "RandomRotationVectorDataset",
    "SAbDabDataset",
    "SKEMPIDataset",
    "SequenceDataset",
    "SizedSequenceDataset",
    "SwissProtDataset",
    "TDCDataset",
    "TherapeuticAntibodyProfilerDataset",
    "Tox21Dataset",
    "ToxCastDataset",
    "TrEMBLDataset",
    "TrajectoryDataset",
    "USPTODataset",
    "USPTOReactionProductDataset",
    "UniProtDataset",
    "UniRef100Dataset",
    "UniRef50Dataset",
    "UniRef90Dataset",
    "VeithCytochromeP4501A2InhibitionDataset",
    "VeithCytochromeP4502C19InhibitionDataset",
    "VeithCytochromeP4502C9InhibitionDataset",
    "VeithCytochromeP4502D6InhibitionDataset",
    "VeithCytochromeP4503A4InhibitionDataset",
    "WangEffectivePermeabilityDataset",
    "ZINCDataset",
    "ZhuAcuteToxicityLD50Dataset",
]



import pytest
from beignet.transforms import Transform


class TestTransform:
    def test___init__(self):
        assert Transform()

    def test__check_inputs(self):
        with pytest.raises(NotImplementedError) as _:
            Transform()._check_inputs([])

    def test__get_params(self):
        assert True

    def test__transform(self):
        with pytest.raises(NotImplementedError) as _:
            Transform()._transform(None, {})

    def test__transformables(self):
        assert True

    def test_extra_repr(self):
        assert True

    def test_forward(self):
        assert True



import beignet
import hypothesis.strategies
import torch.testing
from scipy.spatial.transform import Rotation


@hypothesis.strategies.composite
def _strategy(function):
    axes = function(
        hypothesis.strategies.sampled_from(
            [
                "xyz",
                "xzy",
                "yxz",
                "yzx",
                "zxy",
                "zyx",
                "XYZ",
                "XZY",
                "YXZ",
                "YZX",
                "ZXY",
                "ZYX",
            ]
        ),
    )

    degrees = function(hypothesis.strategies.booleans())

    rotations = Rotation.random(
        function(
            hypothesis.strategies.integers(
                min_value=1,
                max_value=8,
            ),
        ),
    )

    return (
        {
            "input": torch.from_numpy(rotations.as_euler(axes, degrees)),
            "axes": axes,
            "degrees": degrees,
        },
        torch.from_numpy(rotations.inv().as_euler(axes, degrees)),
    )


@hypothesis.given(_strategy())
def test_invert_euler_angle(data):
    parameters, expected = data

    torch.testing.assert_close(
        beignet.invert_euler_angle(**parameters),
        expected,
    )



import math
from typing import Tuple

import torch
from torch import Tensor


def chebyshev_gauss_quadrature(degree: int) -> Tuple[Tensor, Tensor]:
    if not degree > 0:
        raise ValueError

    output = torch.cos(torch.arange(1, 2 * degree, 2) / (2 * degree) * math.pi)

    weight = torch.ones(degree) * (math.pi / degree)

    return output, weight



import torch
from torch import Tensor


def quaternion_to_rotation_vector(
    input: Tensor,
    degrees: bool | None = False,
) -> Tensor:
    r"""
    Convert rotation quaternions to rotation vectors.

    Parameters
    ----------
    input : Tensor, shape=(..., 4)
        Rotation quaternions. Rotation quaternions are normalized to unit norm.

    degrees : bool, optional

    Returns
    -------
    output : Tensor, shape=(..., 3)
        Rotation vectors.
    """
    output = torch.empty(
        [input.shape[0], 3],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    for j in range(input.shape[0]):
        a = input[j, 0]
        b = input[j, 1]
        c = input[j, 2]
        d = input[j, 3]

        if d == 0 and (a == 0 and (b == 0 and c < 0 or b < 0) or a < 0) or d < 0:
            input[j] = -input[j]

        t = input[j, 0] ** 2.0
        u = input[j, 1] ** 2.0
        v = input[j, 2] ** 2.0
        w = input[j, 3] ** 1.0

        y = 2.0 * torch.atan2(torch.sqrt(t + u + v), w)

        if y < 0.001:
            y = 2.0 + y**2.0 / 12 + 7 * y**2.0 * y**2.0 / 2880
        else:
            y = y / torch.sin(y / 2.0)

        output[j] = input[j, :-1] * y

    if degrees:
        output = torch.rad2deg(output)

    return output



import operator
import threading
from tempfile import NamedTemporaryFile
from typing import Callable
from unittest.mock import MagicMock

from beignet.io import ThreadSafeFile


class TestThreadSafeFile:
    def test___init__(self):
        with NamedTemporaryFile() as temporary_file:
            thread_safe_file = ThreadSafeFile(temporary_file.name)

            assert thread_safe_file.path == temporary_file.name

            assert isinstance(thread_safe_file.open, Callable)

            assert isinstance(thread_safe_file.close, Callable)

            assert isinstance(thread_safe_file.storage, threading.local)

    def test___del__(self):
        with NamedTemporaryFile() as temporary_file:
            close = MagicMock()

            thread_safe_file = ThreadSafeFile(temporary_file.name, close=close)

            thread_safe_file.storage.file = MagicMock()

            del thread_safe_file

            close.assert_called_once()

    def test___getattr__(self):
        with NamedTemporaryFile() as temporary_file:
            file = MagicMock()

            thread_safe_file = ThreadSafeFile(temporary_file.name)

            thread_safe_file.storage.file = file

            assert thread_safe_file.read == file.read
            assert thread_safe_file.write == file.write

    def test___getstate__(self):
        with NamedTemporaryFile() as temporary_file:
            thread_safe_file = ThreadSafeFile(temporary_file.name)

            state = thread_safe_file.__getstate__()

            assert "path" in state

            assert "open" in state

            assert "close" in state

            assert "storage" not in state

    def test___setstate__(self):
        with NamedTemporaryFile() as temporary_file:
            thread_safe_file = ThreadSafeFile(temporary_file.name)

            thread_safe_file.__setstate__(
                {
                    "path": temporary_file.name,
                    "open": operator.methodcaller("open"),
                    "close": operator.methodcaller("close"),
                },
            )

            assert thread_safe_file.path == temporary_file.name

            assert isinstance(thread_safe_file.open, Callable)

            assert isinstance(thread_safe_file.close, Callable)

            assert isinstance(thread_safe_file.storage, threading.local)

    def test_file(self):
        with NamedTemporaryFile() as temporary_file:
            open = MagicMock(return_value=MagicMock())

            thread_safe_file = ThreadSafeFile(
                temporary_file.name,
                open=open,
            )

            file = thread_safe_file.file

            open.assert_called_once_with(temporary_file.name)

            assert thread_safe_file.storage.file == file
            assert thread_safe_file.file == file

            assert thread_safe_file.file == file
            assert open.call_count == 1



import torch
from torch import Tensor


def invert_transform(transform: Tensor) -> Tensor:
    """
    Calculates the inverse of an affine transformation matrix.

    Parameters
    ----------
    transform : Tensor
        The affine transformation matrix to be inverted.

    Returns
    -------
    Tensor
        The inverse of the given affine transformation matrix.
    """
    if transform.ndim in {0, 1}:
        return 1.0 / transform

    if transform.ndim == 2:
        return torch.linalg.inv(transform)

    raise ValueError



import beignet
import torch
import torch.testing


def test__lennard_jones_potential():
    x = beignet.lennard_jones_potential(
        torch.tensor(
            [
                [1.0, 2.0],
                [3.0, 4.0],
            ],
        ),
        1.0,
        1.0,
    )

    print(x)

    torch.testing.assert_close(
        beignet.lennard_jones_potential(
            torch.tensor(
                [
                    [1.0, 2.0],
                    [3.0, 4.0],
                ],
            ),
            1.0,
            1.0,
        ),
        torch.tensor(
            [
                [+0.0000000000000000, -0.0615234375000000],
                [-0.0054794428870082, -0.0009763240814209],
            ],
        ),
    )



from torch import Tensor

from ._euler_angle_to_quaternion import (
    euler_angle_to_quaternion,
)
from ._invert_quaternion import invert_quaternion
from ._quaternion_to_euler_angle import (
    quaternion_to_euler_angle,
)


def invert_euler_angle(
    input: Tensor,
    axes: str,
    degrees: bool | None = False,
) -> Tensor:
    r"""
    Invert Euler angles.

    Parameters
    ----------
    input : Tensor, shape (..., 3)
        Euler angles.

    axes : str
        Axes. 1-3 characters belonging to the set {‘X’, ‘Y’, ‘Z’} for intrinsic
        rotations, or {‘x’, ‘y’, ‘z’} for extrinsic rotations. Extrinsic and
        intrinsic rotations cannot be mixed.

    degrees : bool, optional
        If `True`, Euler angles are assumed to be in degrees. Default, `False`.

    Returns
    -------
    inverted_euler_angles : Tensor, shape (..., 3)
        Inverted Euler angles.
    """
    return quaternion_to_euler_angle(
        invert_quaternion(
            euler_angle_to_quaternion(
                input,
                axes,
                degrees,
            ),
        ),
        axes,
        degrees,
    )



import copy

import pytest
import torch
from beignet.features import Feature


class TestFeature:
    def test___deepcopy__(self):
        feature = Feature(torch.tensor([1, 2, 3]))

        with pytest.raises(NotImplementedError):
            copy.deepcopy(feature)

    def test___torch_function__(self):
        feature = Feature(torch.tensor([1, 2, 3]))

        result = feature.__torch_function__(
            torch.add,
            (Feature, torch.Tensor),
            (feature, torch.tensor([1, 2, 3])),
        )

        assert not isinstance(result, Feature)

    def test__to_tensor(self):
        feature = Feature._to_tensor([1, 2, 3])

        assert torch.is_tensor(feature)

        assert not feature.requires_grad

    def test_device(self):
        feature = Feature(torch.tensor([1, 2, 3]))

        assert feature.device == feature.device

    def test_dtype(self):
        assert Feature(torch.tensor([1, 2, 3])).dtype == torch.int64

    def test_ndim(self):
        assert Feature(torch.tensor([1, 2, 3])).ndim == 1

    def test_shape(self):
        assert Feature(torch.tensor([1, 2, 3])).shape == (3,)

    def test_wrap_like(self):
        with pytest.raises(NotImplementedError):
            feature = Feature(torch.tensor([1, 2, 3]))

            Feature.wrap_like(feature, torch.tensor([1, 2, 3]))



import pytest
import torch
from beignet.features import Quaternion
from torch import Size


class TestQuaternion:
    @pytest.fixture
    def rotation_quaternion(self):
        return Quaternion(torch.tensor([0, 1, 0, 0]))

    def test___new__(self):
        result = Quaternion(torch.tensor([0, 1, 0, 0]))

        assert isinstance(result, Quaternion)

        assert result.shape == Size([1, 4])

        with pytest.raises(ValueError):
            Quaternion(torch.tensor([1, 2, 3]))

    def test___repr__(self, rotation_quaternion):
        assert isinstance(rotation_quaternion.__repr__(), str)

    def test__wrap(self):
        result = Quaternion._wrap(torch.tensor([0, 1, 0, 0]))

        assert isinstance(result, Quaternion)

        assert result.shape == Size([4])

    def test_wrap_like(self, rotation_quaternion):
        result = Quaternion.wrap_like(
            rotation_quaternion,
            torch.tensor([1, 0, 0, 0]),
        )

        assert isinstance(result, Quaternion)

        assert result.shape == Size([4])



import math

import torch
import torchaudio.functional
from torch import Tensor

from ._add_chebyshev_polynomial import add_chebyshev_polynomial


def chebyshev_polynomial_power(
    input: Tensor,
    exponent: float | Tensor,
    maximum_exponent: float | Tensor = 16.0,
) -> Tensor:
    input = torch.atleast_1d(input)

    _exponent = int(exponent)

    if _exponent != exponent or _exponent < 0:
        raise ValueError

    if maximum_exponent is not None and _exponent > maximum_exponent:
        raise ValueError

    match _exponent:
        case 0:
            output = torch.tensor([1.0], dtype=input.dtype)
        case 1:
            output = input
        case _:
            output = torch.zeros(input.shape[0] * exponent, dtype=input.dtype)

            output = add_chebyshev_polynomial(output, input)

            index1 = math.prod(input.shape)
            output2 = torch.zeros(2 * index1 - 1, dtype=input.dtype)
            output2[index1 - 1 :] = input / 2.0
            output2 = torch.flip(output2, dims=[0]) + output2
            zs = output2

            index = math.prod(output.shape)
            output1 = torch.zeros(2 * index - 1, dtype=output.dtype)
            output1[index - 1 :] = output / 2.0
            output1 = torch.flip(output1, dims=[0]) + output1
            output = output1

            for _ in range(2, _exponent + 1):
                output = torchaudio.functional.convolve(output, zs, mode="same")

            n = (math.prod(output.shape) + 1) // 2
            c = output[n - 1 :]
            c[1:n] = c[1:n] * 2.0
            output = c

    return output



from torch import Tensor

from ._error_erfc import error_erfc


def error_erf(input: Tensor, *, out: Tensor | None = None) -> Tensor:
    r"""
    Error function.

    Parameters
    ----------
    input : Tensor
        Input tensor.

    out : Tensor, optional
        Output tensor.

    Returns
    -------
    Tensor
    """
    output = 1.0 - error_erfc(input)

    if out is not None:
        out.copy_(output)

        return out

    return output



from typing import Callable

from torch import Tensor
from torch.utils.data import Dataset

from beignet.transforms import Transform


class RandomRotationDataset(Dataset):
    def __init__(
        self,
        data: Tensor,
        *,
        transform: Callable | Transform | None = None,
    ):
        super().__init__()

        self.data = data

        self.transform = transform

    def __getitem__(self, index: int) -> Tensor:
        x = self.data[index]

        if self.transform:
            x = self.transform(x)

        return x

    def __len__(self) -> int:
        return len(self.data)



import functools

import beignet
import torch


def test_polynomial_power():
    for i in range(5):
        for j in range(5):
            torch.testing.assert_close(
                beignet.trim_polynomial_coefficients(
                    beignet.polynomial_power(
                        torch.arange(0.0, i + 1),
                        j,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_polynomial_coefficients(
                    functools.reduce(
                        beignet.multiply_polynomial,
                        [torch.arange(0.0, i + 1)] * j,
                        torch.tensor([1.0]),
                    ),
                    tol=0.000001,
                ),
            )



import beignet
import torch


def test_chebyshev_polynomial_vandermonde():
    v = beignet.chebyshev_polynomial_vandermonde(
        torch.arange(3),
        degree=torch.tensor([3]),
    )

    assert v.shape == (3, 4)

    for i in range(4):
        torch.testing.assert_close(
            v[..., i],
            beignet.evaluate_chebyshev_polynomial(
                torch.arange(3),
                torch.tensor([0.0] * i + [1.0]),
            ),
        )

    v = beignet.chebyshev_polynomial_vandermonde(
        torch.tensor([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]),
        degree=torch.tensor([3]),
    )

    assert v.shape == (3, 2, 4)

    for i in range(4):
        torch.testing.assert_close(
            v[..., i],
            beignet.evaluate_chebyshev_polynomial(
                torch.tensor([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]),
                torch.tensor([0.0] * i + [1.0]),
            ),
        )



import beignet
import hypothesis.strategies
from beignet.datasets import RandomEulerAngleDataset


@hypothesis.strategies.composite
def _strategy(function):
    size = function(
        hypothesis.strategies.integers(
            min_value=1,
            max_value=8,
        ),
    )

    axes = function(
        hypothesis.strategies.sampled_from(
            [
                "xyz",
                "xzy",
                "yxz",
                "yzx",
                "zxy",
                "zyx",
                "XYZ",
                "XZY",
                "YXZ",
                "YZX",
                "ZXY",
                "ZYX",
            ]
        ),
    )

    degrees = function(hypothesis.strategies.booleans())

    return (
        {
            "size": size,
            "axes": axes,
            "degrees": degrees,
        },
        beignet.random_euler_angle(size, axes=axes, degrees=degrees),
    )


class TestRandomEulerAngleDataset:
    @hypothesis.given(_strategy())
    def test___init__(self, data):
        parameters, output = data

        dataset = RandomEulerAngleDataset(**parameters)

        assert dataset.data.shape == output.shape

        assert dataset.data.dtype == output.dtype

        assert dataset.data.layout == output.layout



import torch
from torch import Tensor


def legendre_polynomial_companion(input: Tensor) -> Tensor:
    input = torch.atleast_1d(input)

    if input.shape[0] < 2:
        raise ValueError

    if input.shape[0] == 2:
        return torch.tensor([[-input[0] / input[1]]])

    n = input.shape[0] - 1

    output = torch.zeros((n, n), dtype=input.dtype)

    scale = 1.0 / torch.sqrt(2 * torch.arange(n) + 1)

    shape = output.shape

    output = torch.reshape(output, [-1])

    output[1 :: n + 1] = torch.arange(1, n) * scale[: n - 1] * scale[1:n]

    output[n :: n + 1] = torch.arange(1, n) * scale[: n - 1] * scale[1:n]

    output = torch.reshape(output, shape)

    output[:, -1] += -(input[:-1] / input[-1]) * (scale / scale[-1]) * (n / (2 * n - 1))

    return output



import json
from gzip import GzipFile
from io import BytesIO
from pathlib import Path
from typing import Any, Callable, Dict, Union

from torch.utils.data import Dataset

from beignet.transforms import Transform

try:
    import lmdb
except ImportError as error:
    raise ImportError(
        """
        LMDB datasets require the `lmdb` dependency:

            $ pip install "beignet[lmdb]"
        """
    ) from error


class LMDBDataset(Dataset):
    def __init__(
        self,
        root: Union[str, Path],
        *,
        lock: bool = False,
        max_readers: int = 1,
        meminit: bool = True,
        readahead: bool = True,
        readonly: bool = True,
        transform: Union[Callable, Transform, None] = None,
    ):
        super().__init__()

        self._root = root

        self._transform_fn = transform

        if isinstance(self._root, str):
            self._root = Path(self._root).resolve()

        self._data = lmdb.open(
            str(self._root),
            lock=lock,
            max_readers=max_readers,
            meminit=meminit,
            readahead=readahead,
            readonly=readonly,
        )

        with self._data.begin(write=False) as transaction:
            self._size = int(transaction.get(b"num_examples"))

    def __getitem__(self, index: int) -> Dict[str, Any]:
        if not 0 <= index < self._size:
            raise IndexError(index)

        with self._data.begin(write=False) as transaction:
            with GzipFile(
                fileobj=BytesIO(transaction.get(str(index).encode())),
                mode="rb",
            ) as descriptor:
                item = json.loads(descriptor.read())

        if self._transform_fn:
            item = self._transform_fn(item)

        return item

    def __len__(self) -> int:
        return self._size



from beignet.datasets import SizedSequenceDataset


class TestSizedSequenceDataset:
    def test___init__(self, mocker):
        mock_sequence_dataset_init = mocker.patch(
            "beignet.datasets.SequenceDataset.__init__",
            return_value=None,
        )

        sizes = [1, 2, 3]

        dataset = SizedSequenceDataset("/foo/bar", sizes)

        mock_sequence_dataset_init.assert_called_once_with("/foo/bar")

        assert dataset.sizes == sizes

    def test___len__(self):
        sizes = [1, 2, 3]

        assert len(SizedSequenceDataset("/foo/bar", sizes)) == len(sizes)



import math

import torch
from torch import Tensor

from ._linear_polynomial import linear_polynomial
from ._multiply_polynomial import multiply_polynomial


def polynomial_from_roots(input: Tensor) -> Tensor:
    r"""
    Parameters
    ----------
    input : Tensor
        Roots.

    Returns
    -------
    output : Tensor
        Polynomial coefficients.
    """
    f = linear_polynomial
    g = multiply_polynomial
    if math.prod(input.shape) == 0:
        return torch.ones([1])

    input, _ = torch.sort(input)

    ys = []

    for x in input:
        a = torch.zeros(input.shape[0] + 1, dtype=x.dtype)
        b = f(-x, 1)

        a = torch.atleast_1d(a)
        b = torch.atleast_1d(b)

        dtype = torch.promote_types(a.dtype, b.dtype)

        a = a.to(dtype)
        b = b.to(dtype)

        if a.shape[0] > b.shape[0]:
            y = torch.concatenate(
                [
                    b,
                    torch.zeros(
                        a.shape[0] - b.shape[0],
                        dtype=b.dtype,
                    ),
                ],
            )

            y = a + y
        else:
            y = torch.concatenate(
                [
                    a,
                    torch.zeros(
                        b.shape[0] - a.shape[0],
                        dtype=a.dtype,
                    ),
                ]
            )

            y = b + y

        ys = [*ys, y]

    p = torch.stack(ys)

    m = p.shape[0]

    x = m, p

    while x[0] > 1:
        m, r = divmod(x[0], 2)

        z = x[1]

        previous = torch.zeros([len(p), input.shape[0] + 1])

        y = previous

        for i in range(0, m):
            y[i] = g(z[i], z[i + m])[: input.shape[0] + 1]

        previous = y

        if r:
            previous[0] = g(previous[0], z[2 * m])[: input.shape[0] + 1]

        x = m, previous

    _, output = x

    return output[0]



# import beignet.special
# import hypothesis
# import hypothesis.strategies
# import scipy
# import torch
#
#
# @hypothesis.strategies.composite
# def _strategy(function):
#     x, y = torch.meshgrid(
#         torch.linspace(
#             function(
#                 hypothesis.strategies.floats(
#                     min_value=-10,
#                     max_value=-10,
#                 ),
#             ),
#             function(
#                 hypothesis.strategies.floats(
#                     min_value=10,
#                     max_value=10,
#                 ),
#             ),
#             steps=128,
#             dtype=torch.float64,
#         ),
#         torch.linspace(
#             function(
#                 hypothesis.strategies.floats(
#                     min_value=-10,
#                     max_value=-10,
#                 ),
#             ),
#             function(
#                 hypothesis.strategies.floats(
#                     min_value=10,
#                     max_value=10,
#                 ),
#             ),
#             steps=128,
#             dtype=torch.float64,
#         ),
#         indexing="xy",
#     )
#
#     input = x + 1.0j * y
#
#     return input, scipy.special.dawsn(input)
#
#
# @hypothesis.given(_strategy())
# def test_dawson_integral_f(data):
#     input, output = data
#
#     torch.testing.assert_close(beignet.special.dawson_integral_f(input), output)



import beignet
import torch


def test_probabilists_hermite_polynomial_vandermonde():
    x = torch.arange(3)
    v = beignet.probabilists_hermite_polynomial_vandermonde(
        x,
        3,
    )
    assert v.shape == (3, 4)
    for i in range(4):
        coefficients = torch.tensor([0.0] * i + [1.0])
        torch.testing.assert_close(
            v[..., i],
            beignet.evaluate_probabilists_hermite_polynomial(x, coefficients),
        )

    x = torch.tensor([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
    v = beignet.probabilists_hermite_polynomial_vandermonde(
        x,
        3,
    )
    assert v.shape == (3, 2, 4)
    for i in range(4):
        coefficients = torch.tensor([0.0] * i + [1.0])
        torch.testing.assert_close(
            v[..., i],
            beignet.evaluate_probabilists_hermite_polynomial(x, coefficients),
        )



import torch
from torch import Tensor


def rotation_vector_to_quaternion(
    input: Tensor,
    degrees: bool | None = False,
    canonical: bool | None = False,
) -> Tensor:
    r"""
    Convert rotation vector to rotation quaternion.

    Parameters
    ----------
    input : Tensor, shape=(..., 3)
        Rotation vector.

    degrees : bool, optional
        If `True`, rotation vector magnitudes are assumed to be in degrees.
        Default, `False`.

    canonical : bool, optional
        Whether to map the redundant double cover of rotation space to a unique
        canonical single cover. If `True`, then the rotation quaternion is
        chosen from :math:`{q, -q}` such that the :math:`w` term is positive.
        If the :math:`w` term is :math:`0`, then the rotation quaternion is
        chosen such that the first non-zero term of the :math:`x`, :math:`y`,
        and :math:`z` terms is positive.

    Returns
    -------
    output : Tensor, shape=(..., 4)
        Rotation quaternion.
    """
    if degrees:
        input = torch.deg2rad(input)

    output = torch.empty(
        [input.shape[0], 4],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    for j in range(input.shape[0]):
        t = input[j, 0] ** 2.0
        u = input[j, 1] ** 2.0
        v = input[j, 2] ** 2.0

        y = torch.sqrt(t + u + v)

        if y < 0.001:
            scale = 0.5 - y**2.0 / 48.0 + y**2.0 * y**2.0 / 3840.0
        else:
            scale = torch.sin(y / 2) / y

        output[j, :-1] = input[j] * scale

        output[j, 3] = torch.cos(y / 2)

    if canonical:
        for j in range(output.shape[0]):
            a = output[j, 0]
            b = output[j, 1]
            c = output[j, 2]
            d = output[j, 3]

            if d == 0 and (a == 0 & (b == 0 & c < 0 | b < 0) | a < 0) | d < 0:
                output[j] = -output[j]

    return output



import math
from typing import Literal

import torch
import torchaudio.functional
from torch import Tensor


def multiply_chebyshev_polynomial(
    input: Tensor,
    other: Tensor,
    mode: Literal["full", "same", "valid"] = "full",
) -> Tensor:
    r"""
    Returns the product of two polynomials.

    Parameters
    ----------
    input : Tensor
        Polynomial coefficients.

    other : Tensor
        Polynomial coefficients.

    Returns
    -------
    output : Tensor
        Polynomial coefficients of the product.
    """
    input = torch.atleast_1d(input)
    other = torch.atleast_1d(other)

    dtype = torch.promote_types(input.dtype, other.dtype)

    input = input.to(dtype)
    other = other.to(dtype)

    index = math.prod(input.shape)
    output1 = torch.zeros(2 * index - 1, dtype=input.dtype)
    output1[index - 1 :] = input / 2.0
    output1 = torch.flip(output1, dims=[0]) + output1
    a = output1

    index1 = math.prod(other.shape)
    output2 = torch.zeros(2 * index1 - 1, dtype=other.dtype)
    output2[index1 - 1 :] = other / 2.0
    output2 = torch.flip(output2, dims=[0]) + output2
    b = output2

    output = torchaudio.functional.convolve(a, b, mode=mode)

    n = (math.prod(output.shape) + 1) // 2
    c = output[n - 1 :]
    c[1:n] = c[1:n] * 2.0
    output = c

    if mode == "same":
        output = output[: max(input.shape[0], other.shape[0])]

    return output



from os import PathLike
from pathlib import Path
from typing import Callable

import pooch
from pooch import Decompress

from beignet.transforms import Transform

from ._fasta_dataset import FASTADataset


class UniProtDataset(FASTADataset):
    def __init__(
        self,
        url: str,
        root: str | PathLike | None = None,
        known_hash: str | None = None,
        *,
        transform: Callable | Transform | None = None,
        target_transform: Callable | Transform | None = None,
    ):
        """
        Parameters
        ----------
        url : str
            URL to the file that needs to be downloaded. Ideally, the URL
            should end with a file name (e.g., `uniref50.fasta.gz`).

        root : str | PathLike, optional
            Root directory where the dataset subdirectory exists or, if
            `download` is `True`, the directory where the dataset subdirectory
            will be created and the dataset downloaded.

        transform : Callable | Transform, optional
            A `Callable` or `Transform` that that maps a sequence to a
            transformed sequence (default: `None`).

        target_transform : Callable | Transform, optional
            A `Callable` or `Transform` that maps a target (a cluster
            identifier) to a transformed target (default: `None`).
        """
        if root is None:
            root = pooch.os_cache("beignet")

        if isinstance(root, str):
            root = Path(root)

        self.root = root.resolve()

        name = self.__class__.__name__.replace("Dataset", "")

        super().__init__(
            pooch.retrieve(
                url,
                known_hash,
                f"{name}.fasta.gz",
                root / name,
                processor=Decompress(
                    name=f"{name}.fasta",
                ),
                progressbar=True,
            ),
        )

        self.transform = transform

        self.target_transform = target_transform

    def __getitem__(self, index: int) -> (str, str):
        input, target = self.get(index)

        if self.transform:
            input = self.transform(input)

        if self.target_transform:
            target = self.target_transform(target)

        return input, target



import beignet
import hypothesis.strategies
import torch.testing
from scipy.spatial.transform import Rotation


@hypothesis.strategies.composite
def _strategy(function):
    rotation = Rotation.random(
        function(
            hypothesis.strategies.integers(
                min_value=1,
                max_value=8,
            ),
        ),
    )

    return (
        {
            "input": torch.from_numpy(
                rotation.as_quat(
                    canonical=False,
                ),
            ),
        },
        torch.unsqueeze(
            torch.abs(
                torch.from_numpy(
                    rotation.mean().as_quat(
                        canonical=False,
                    ),
                ),
            ),
            dim=0,
        ),
    )


@hypothesis.given(_strategy())
def test_quaternion_mean(data):
    parameters, expected = data

    torch.testing.assert_close(
        torch.abs(
            beignet.quaternion_mean(
                **parameters,
            ),
        ),
        expected,
    )



from torch import Tensor


def invert_rotation_vector(
    input: Tensor,
) -> Tensor:
    r"""
    Invert rotation vectors.

    Parameters
    ----------
    input : Tensor, shape (..., 3)
        Rotation vectors.

    Returns
    -------
    inverted_rotation_vectors : Tensor, shape (..., 3)
        Inverted rotation vectors.
    """
    return -input



from __future__ import annotations

from types import ModuleType
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import torch
from torch import Tensor
from torch._C import DisableTorchFunctionSubclass
from torch.types import _device, _dtype, _size

F = TypeVar("F", bound="Feature")


class Feature(Tensor):
    """
    Feature
    """

    __f: Optional[ModuleType] = None

    @staticmethod
    def _to_tensor(
        data: Any,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[torch.device, str, int]] = None,
        requires_grad: Optional[bool] = None,
    ) -> Tensor:
        if requires_grad is None:
            if isinstance(data, Tensor):
                requires_grad = data.requires_grad
            else:
                requires_grad = False

        tensor = torch.as_tensor(data, dtype=dtype, device=device)

        return tensor.requires_grad_(requires_grad)

    @classmethod
    def wrap_like(cls: Type[F], other: F, tensor: Tensor) -> F:
        raise NotImplementedError

    # NOTE:
    #   We don’t need to wrap the output of ``tensor.requires_grad_``,
    #   because it’s an inplace operation and automatically retains its type.
    _NO_WRAPPING_EXCEPTIONS = {
        Tensor.clone: lambda cls, input, output: cls.wrap_like(input, output),
        Tensor.detach: lambda cls, input, output: cls.wrap_like(input, output),
        Tensor.requires_grad_: lambda cls, input, output: output,
        Tensor.to: lambda cls, input, output: cls.wrap_like(input, output),
    }

    @classmethod
    def __torch_function__(
        cls,
        func: Callable[..., Tensor],
        types: Tuple[Type[Tensor], ...],
        args: Sequence[Any] = (),
        kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Tensor:
        """
        The default behavior of :class:`~Tensor`’s retains the custom tensor
        type. For :class:`Feature`, this creates two problems:

            1.  :class:`Feature` may require metadata and the default wrapping,
                i.e., ``return cls(func(*args, **kwargs))``, will fail.

            2.  For most operations, there is no way of knowing if an input
                type is still valid for the output type.

        To address these two problems, :class:`Feature` disables automatic
        output wrapping for most operators. The exceptions are available from
        :attr:`Feature._NO_WRAPPING_EXCEPTIONS`
        """
        # NOTE:
        #   ``super().__torch_function__`` has no hook to prevent the
        #   coercing of the output type into the input type so this
        #   functionality is reimplemented.
        if not all(issubclass(cls, t) for t in types):
            return NotImplemented

        with DisableTorchFunctionSubclass():
            output = func(*args, **kwargs or dict())

            wrapper = cls._NO_WRAPPING_EXCEPTIONS.get(func)

            # NOTE:
            #   Besides ``func`` being an exception, this class method
            #   requires the input operand, i.e. ``args[0]``, to be an
            #   instance of the class that ``__torch_function__`` invoked.
            #   The ``__torch_function__`` protocol will invoke this method
            #   on each type in the computation by walking the method
            #   resolution  order (MRO). For example,
            #   ``Tensor(...).to(beignet.features.Foo( ...))`` invokes
            #   ``beignet.features.Foo.__torch_function__`` with
            #   ``args = (Tensor(), beignet.featues.Foo())``.
            #   Without this guard, ``Tensor`` would be wrapped into
            #   ``beignet.features.Foo``.
            if wrapper and isinstance(args[0], cls):
                return wrapper(cls, args[0], output)

            # NOTE:
            #   Because inplace ``func``’s, canonically identified with a
            #   trailing underscore in their name, e.g., ``.add_(...)``,
            #   retain their input type, they need to be unwrapped.
            if isinstance(output, cls):
                return output.as_subclass(Tensor)

            return output

    def _make_repr(self, **kwargs: Any) -> str:
        items = []

        for key, value in kwargs.items():
            items.append(f"{key}={value}")

        return f"{super().__repr__()[:-1]}, {', '.join(items)})"

    @property
    def _f(self) -> ModuleType:
        # NOTE:
        #   Lazy import of ``beignet.transforms.functional`` to bypass the
        #   ``ImportError`` raised by the circual import. The
        #   ``beignet.transforms.functional`` import is deferred until the
        #   functional module is referenced and it’s shared across all
        #   instances of the class.
        if Feature.__f is None:
            import beignet.transforms.functional

            Feature.__f = beignet.transforms.functional

        return Feature.__f

    @property
    def device(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> _device:  # type: ignore[override]
        with DisableTorchFunctionSubclass():
            return super().device

    @property
    def ndim(self) -> int:  # type: ignore[override]
        with DisableTorchFunctionSubclass():
            return super().ndim

    @property
    def dtype(self) -> _dtype:  # type: ignore[override]
        with DisableTorchFunctionSubclass():
            return super().dtype

    @property
    def shape(self) -> _size:  # type: ignore[override]
        with DisableTorchFunctionSubclass():
            return super().shape

    def __deepcopy__(self: F, memo: Dict[int, Any]) -> F:
        # NOTE:
        #   Detach, because ``deepcopy(Tensor)``, unlike ``Tensor.clone``,
        #   isn’t be added to the computational graph.

        # NOTE:
        #   Because a side-effect of detaching is clearing
        #   ``Tensor.requires_grad``, it’s refilled before returning.

        # NOTE:
        #   Deep-copying of metadata isn’t explicitly handled.
        return (
            self.detach()
            .clone()
            .requires_grad_(
                self.requires_grad,
            )
        )  # type: ignore[return-value]


_InputType = Union[
    Dict[str, Tensor],
    Feature,
    Sequence[Tensor],
    Tensor,
    str,
]

_InputTypeJIT = Union[Dict[str, Tensor], Sequence[Tensor], Tensor]



import beignet
import hypothesis.strategies


@hypothesis.strategies.composite
def _strategy(function):
    size = function(
        hypothesis.strategies.integers(
            min_value=1,
            max_value=8,
        ),
    )

    canonical = function(hypothesis.strategies.booleans())

    return (
        {
            "size": size,
            "canonical": canonical,
        },
        None,
    )


@hypothesis.given(_strategy())
def test_random_quaternion(data):
    parameters, _ = data

    assert beignet.random_quaternion(
        **parameters,
    ).shape == (parameters["size"], 4)



# import beignet
# import pytest
# import torch
#
#
# def test_differentiate_laguerre_polynomial():
#     with pytest.raises(TypeError):
#         beignet.differentiate_laguerre_polynomial(
#             torch.tensor([0]),
#             order=0.5,
#         )
#
#     with pytest.raises(ValueError):
#         beignet.differentiate_laguerre_polynomial(
#             torch.tensor([0]),
#             order=-1,
#         )
#
#     for i in range(5):
#         torch.testing.assert_close(
#             beignet.trim_laguerre_polynomial_coefficients(
#                 beignet.differentiate_laguerre_polynomial(
#                     torch.tensor([0.0] * i + [1.0]),
#                     order=0,
#                 ),
#                 tol=0.000001,
#             ),
#             beignet.trim_laguerre_polynomial_coefficients(
#                 torch.tensor([0.0] * i + [1.0]),
#                 tol=0.000001,
#             ),
#         )
#
#     for i in range(5):
#         for j in range(2, 5):
#             torch.testing.assert_close(
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     beignet.differentiate_laguerre_polynomial(
#                         beignet.integrate_laguerre_polynomial(
#                             torch.tensor([0.0] * i + [1.0]),
#                             order=j,
#                         ),
#                         order=j,
#                     ),
#                     tol=0.000001,
#                 ),
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     torch.tensor([0.0] * i + [1.0]),
#                     tol=0.000001,
#                 ),
#             )
#
#     for i in range(5):
#         for j in range(2, 5):
#             torch.testing.assert_close(
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     beignet.differentiate_laguerre_polynomial(
#                         beignet.integrate_laguerre_polynomial(
#                             torch.tensor([0.0] * i + [1.0]),
#                             order=j,
#                             scale=2,
#                         ),
#                         order=j,
#                         scale=0.5,
#                     ),
#                     tol=0.000001,
#                 ),
#                 beignet.trim_laguerre_polynomial_coefficients(
#                     torch.tensor([0.0] * i + [1.0]),
#                     tol=0.000001,
#                 ),
#             )
#
#     # c2d = torch.rand(3, 4)
#
#     # torch.testing.assert_close(
#     #     beignet.lagder(c2d, axis=0),
#     #     torch.vstack([beignet.lagder(c) for c in c2d.T]).T,
#     # )
#
#     # torch.testing.assert_close(
#     #     beignet.lagder(
#     #         c2d,
#     #         axis=1,
#     #     ),
#     #     torch.vstack([beignet.lagder(c) for c in c2d]),
#     # )



import beignet
import torch


def test_laguerre_polynomial_x():
    torch.testing.assert_close(
        beignet.laguerre_polynomial_x,
        torch.tensor([1.0, -1.0]),
        check_dtype=False,
    )



import torch
from torch import Tensor


def invert_rotation_matrix(input: Tensor) -> Tensor:
    r"""
    Invert rotation matrices.

    Parameters
    ----------
    input : Tensor, shape (..., 3, 3)
        Rotation matrices.

    Returns
    -------
    inverted_rotation_matrices : Tensor, shape (..., 3, 3)
        Inverted rotation matrices.
    """
    return torch.transpose(input, -2, -1)



import beignet
import hypothesis.extra.numpy
import hypothesis.strategies
import numpy
import torch.testing
from scipy.spatial.transform import Rotation


@hypothesis.strategies.composite
def _strategy(function):
    size = function(
        hypothesis.strategies.integers(
            min_value=1,
            max_value=8,
        ),
    )

    input = function(
        hypothesis.extra.numpy.arrays(
            numpy.float64,
            (size, 3),
            elements={
                "allow_infinity": False,
                "min_value": numpy.finfo(numpy.float32).min,
                "max_value": numpy.finfo(numpy.float32).max,
            },
        ),
    )

    rotation = Rotation.random(size)

    canonical = function(hypothesis.strategies.booleans())

    inverse = function(hypothesis.strategies.booleans())

    return (
        {
            "input": torch.from_numpy(input),
            "rotation": torch.from_numpy(rotation.as_quat(canonical)),
            "inverse": inverse,
        },
        torch.from_numpy(rotation.apply(input, inverse)),
    )


@hypothesis.given(_strategy())
def test_apply_quaternion(data):
    parameters, expected = data

    torch.testing.assert_close(
        beignet.apply_quaternion(**parameters),
        expected,
    )



import os.path
from pathlib import Path
from typing import Callable, Union

import torch
from Bio.PDB import PDBParser
from torch import Tensor

import beignet.io
from beignet.transforms import Transform

from ._parquet_dataset import ParquetDataset


class SKEMPIDataset(ParquetDataset):
    """
    The Structural Kinetic and Energetic database of Mutant Protein
    Interactions (SKEMPI) database is a compilation of experimental data on
    the thermodynamics of mutations in protein-protein interactions. The
    database includes protein names, protein structures from the Protein
    Data Bank (PDB), mutation information, and the change in free energy
    upon mutation. The change in free energy gives an indication of how the
    mutation affects the binding affinity of the two proteins.
    """

    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = False,
        sequence_transform_fn: Union[Callable, Transform, None] = None,
        structure_transform_fn: Union[Callable, Transform, None] = None,
        target_transform: Union[Callable, Transform, None] = None,
    ) -> None:
        """
        :param root: Root directory where the dataset subdirectory exists or,
            if :attr:`download` is ``True``, the directory where the dataset
            subdirectory will be created and the dataset downloaded.

        :param download: If ``True``, download the dataset to the :attr:`root`
            directory (default: ``False``). If the dataset is already
            downloaded, it is not redownloaded.

        :param sequence_transform_fn: A ``Callable`` or ``Transform`` that maps
            sequences to transformed sequences (default: ``None``).

        :param structure_transform_fn: A ``Callable`` or ``Transform`` that
            maps structures to transformed structures (default: ``None``).

        :param target_transform: ``Callable`` or ``Transform`` that maps a
            target to a transformed target (default: ``None``).
        """
        if isinstance(root, str):
            root = Path(root).resolve()

        self._root = root

        if download:
            beignet.io.download(
                source="s3://beignet-data-dev/designdb/lake/thirdparty/skempi/cc5952a4a37f4f1fbe14ce484a00eb87_0.snappy.parquet",
                destination=self._root / "SKEMPI-v2.0",
                filename="SKEMPI-v2.0.parquet",
            )

            beignet.io.download_and_extract_archive(
                resource="https://life.bsc.es/pid/skempi2/database/download/SKEMPI2_PDBs.tgz",
                source=self._root,
                destination=self._root,
                name="SKEMPI-v2.0.tar.gz",
                remove_archive=True,
            )

        super().__init__(
            self._root / "SKEMPI-v2.0",
            self._root / "SKEMPI-v2.0" / "SKEMPI-v2.0.parquet",
        )

        self._sequence_transform_fn = sequence_transform_fn

        self._structure_transform_fn = structure_transform_fn

        self._target_transform_fn = target_transform

        self._data = self._data.dropna(
            subset=[
                "affinity_antigen_sequence",
                "affinity_pkd",
                "fv_heavy",
                "fv_light",
            ],
        )

        self._parser = PDBParser()

        self._structure_paths = [*self._root.glob("PDBs/*.pdb")]

    def __getitem__(
        self,
        index: int,
    ) -> (((str, str, str), (Tensor, [str])), (float, ...)):
        """
        :param index: index of the record to return.

        :returns: A pair of the form:

            .. math::

                \\left(\\text{antibodies},\\;\\text{targets}\\right).

            Each antibody in :math:`\\text{antibodies}` is a pair of the form:

            .. math::

                \\left(\\text{sequences},\\;\\text{structures}\\right).

            :math:`\\text{sequences} `is a :math:`3`-tuple of the form:

            .. math::

                \\left(\\text{VH},\\;\\text{VL},\\;\\text{Ag}\\right)

            where `\\text{VH}` is a ``str`` that represents the
            immunoglobulin heavy chain variable region sequence, `\\text{
            VL}` is a ``str`` that represents the immunoglobulin light chain
            variable region sequence, and $\\text{Ag}` is a ``str`` that
            represents the antigen sequence.

            An antibody is made up of two heavy chains and two light chains.
            Each heavy and light chain has a variable (:math:`V`) region and
            a constant (:math:`C`) region. The variable regions of the heavy
            and light chains form the antigen-binding site of the antibody.
            Each variable region is unique and gives the antibody its
            specificity for binding to a particular antigen. The heavy and
            light chain variable regions are named for their extensive
            sequence variability among different antibodies. This
            variability allows the immune system to produce antibodies that
            can specifically recognize and bind to a vast array of antigens.

            Antigens are molecules capable of stimulating an immune
            response. They are typically proteins or polysaccharides. This
            includes c omponents of bacterial cell walls, capsules, pili,
            and bacterial flagella, as well as proteins in viruses.

            The immune system recognizes antigens as foreign and mounts an
            immune response against them. Antigens are recognized by
            specific antibodies, which bind to the antigen. This binding can
            neutralize the antigen, mark it for destruction by other immune
            cells, or trigger other types of immune responses. Each type of
            antibody recognizes and binds to a specific antigen; this
            specificity is determined by the variable regions of the
            antibody's heavy and light chains.

            :math:`\\text{pKd}` is the negative logarithm of the
            dissociation constant (:math:`\\text{Kd}`). The dissociation
            constant is a measure of how tightly a ligand (e.g., a drug)
            binds to a receptor. The smaller the ``Kd`` value, the tighter
            or stronger the binding between the ligand and its receptor.
            Because :math:`\\text{pKd}` is the negative logarithm of
            :math:`\\text{Kd}`, a larger :math:`\\text{pKd}` value therefore
            represents stronger binding affinity. The :math:`\\text{pKd}`
            value is commonly used in pharmacology and medicinal chemistry
            because it allows easier comparison of binding affinities across
            different ligand-receptor pairs. It’s an important metric when
            assessing the potential efficacy of a drug.
        """
        item = super().__getitem__(index)

        sequence = (
            item["fv_heavy"],
            item["fv_light"],
            item["affinity_antigen_sequence"],
        )

        if self._sequence_transform_fn is not None:
            sequence = self._sequence_transform_fn(sequence)

        name, _ = os.path.splitext(
            os.path.basename(
                self._structure_paths[index],
            ),
        )

        structure = self._parser.get_structure(
            name,
            self._structure_paths[index],
        )

        atomic_coordinates = []

        residue_names = []

        atom_names = []

        alternate_location_indicators = []

        for atom in [*structure.get_atoms()]:
            atomic_coordinates = [
                *atomic_coordinates,
                torch.from_numpy(atom.coord),
            ]

            (
                _,
                _,
                residue_name,
                atom_name,
                alternate_location_indicator,
            ) = atom.get_full_id()

            residue_names = [
                *residue_names,
                residue_name,
            ]

            atom_names = [*atom_names, atom_name]

            alternate_location_indicator, _ = alternate_location_indicator

            alternate_location_indicators = [
                *alternate_location_indicators,
                alternate_location_indicator,
            ]

        structure = (
            torch.stack(atomic_coordinates),
            residue_names,
        )

        if self._structure_transform_fn is not None:
            structure = self._structure_transform_fn(sequence)

        target = item["affinity_pkd"]

        if self._target_transform_fn is not None:
            target = self._target_transform_fn(sequence)

        return (sequence, structure), target



import beignet
import torch


def test_polynomial_roots():
    torch.testing.assert_close(
        beignet.polynomial_roots(torch.tensor([1.0])),
        torch.tensor([]),
    )

    torch.testing.assert_close(
        beignet.polynomial_roots(torch.tensor([1.0, 2.0])),
        torch.tensor([-0.5]),
    )

    for index in range(2, 5):
        input = torch.linspace(-1, 1, index)

        torch.testing.assert_close(
            beignet.trim_polynomial_coefficients(
                beignet.polynomial_roots(
                    beignet.polynomial_from_roots(
                        input,
                    ),
                ),
                tol=0.000001,
            ),
            beignet.trim_polynomial_coefficients(
                input,
                tol=0.000001,
            ),
        )



import math

import beignet
import torch


def test_chebyshev_gauss_quadrature():
    output, weight = beignet.chebyshev_gauss_quadrature(100)

    output = beignet.chebyshev_polynomial_vandermonde(
        output,
        degree=torch.tensor([99]),
    )

    u = (output.T * weight) @ output

    v = 1 / torch.sqrt(u.diagonal())

    torch.testing.assert_close(
        v[:, None] * u * v,
        torch.eye(100),
    )

    torch.testing.assert_close(
        torch.sum(weight),
        torch.tensor(math.pi),
    )



import torch
from torch import Tensor


def translation_identity(
    size: int,
    *,
    out: Tensor | None = None,
    dtype: torch.dtype | None = None,
    layout: torch.layout | None = torch.strided,
    device: torch.device | None = None,
    requires_grad: bool | None = False,
) -> Tensor:
    """
    Identity translation vectors.

    Parameters
    ----------
    size : int
        Output size.

    out : Tensor, optional
        Output tensor. Default, `None`.

    dtype : torch.dtype, optional
        Type of the returned tensor. Default, global default.

    layout : torch.layout, optional
        Layout of the returned tensor. Default, `torch.strided`.

    device : torch.device, optional
        Device of the returned tensor. Default, current device for the default
        tensor type.

    requires_grad : bool, optional
        Whether autograd records operations on the returned tensor. Default,
        `False`.

    Returns
    -------
    output : Tensor, shape=(size, 3)
        Identity rotation quaternions.
    """
    return torch.zeros(
        [size, 3],
        out=out,
        dtype=dtype,
        layout=layout,
        device=device,
        requires_grad=requires_grad,
    )



import functools
import operator

import torch
from torch import Tensor

from ._laguerre_polynomial_vandermonde import laguerre_polynomial_vandermonde


def laguerre_polynomial_vandermonde_2d(
    x: Tensor,
    y: Tensor,
    degree: Tensor,
) -> Tensor:
    functions = (
        laguerre_polynomial_vandermonde,
        laguerre_polynomial_vandermonde,
    )

    n = len(functions)

    if n != len([x, y]):
        raise ValueError

    if n != len(degree):
        raise ValueError

    if n == 0:
        raise ValueError

    matrices = []

    for i in range(n):
        matrix = functions[i]((x, y)[i], degree[i])

        matrices = [
            *matrices,
            matrix[(..., *tuple(slice(None) if j == i else None for j in range(n)))],
        ]

    vandermonde = functools.reduce(
        operator.mul,
        matrices,
    )

    return torch.reshape(
        vandermonde,
        [*vandermonde.shape[: -len(degree)], -1],
    )



import torch
from torch import Tensor


def apply_rotation_matrix(
    input: Tensor,
    rotation: Tensor,
    inverse: bool | None = False,
) -> Tensor:
    r"""
    Rotates vectors in three-dimensional space using rotation matrices.

    Note
    ----
    This function interprets the rotation of the original frame to the final
    frame as either a projection, where it maps the components of vectors from
    the final frame to the original frame, or as a physical rotation,
    integrating the vectors into the original frame during the rotation
    process. Consequently, the vector components are maintained in the original
    frame’s perspective both before and after the rotation.

    Parameters
    ----------
    input : Tensor, shape (..., 3)
        Each vector represents a vector in three-dimensional space. The number
        of rotation matrices and number of vectors must follow standard
        broadcasting rules: either one of them equals unity or they both equal
        each other.

    rotation : Tensor, shape (..., 3, 3)
        Rotation matrices.

    inverse : bool, optional
        If `True` the inverse of the rotation matrices are applied to the input
        vectors. Default, `False`.

    Returns
    -------
    rotated_vectors : Tensor, shape (..., 3)
        Rotated vectors.
    """
    if inverse:
        output = torch.einsum("ikj, ik -> ij", rotation, input)
    else:
        output = torch.einsum("ijk, ik -> ij", rotation, input)

    return output



from pathlib import Path
from typing import Callable, List, Tuple, TypeVar

import pandas
import pooch
from torch.utils.data import Dataset

from beignet.transforms import Transform

T = TypeVar("T")


class TDCDataset(Dataset):
    _x: List[T]
    _y: List[T]

    def __init__(
        self,
        root: str | Path,
        download: bool = False,
        *,
        identifier: int,
        suffix: str,
        checksum: str,
        x_keys: List[str],
        y_keys: List[str] | None = None,
        transform: Callable | Transform | None = None,
        target_transform: Callable | Transform | None = None,
    ):
        super().__init__()

        if isinstance(root, str):
            root = Path(root)

        if download:
            pooch.retrieve(
                f"https://dataverse.harvard.edu/api/access/datafile/{identifier}",
                fname=f"{self.__class__.__name__}.{suffix}",
                known_hash=checksum,
                path=root / self.__class__.__name__,
                progressbar=True,
            )

        path = root / self.__class__.__name__ / f"{self.__class__.__name__}.{suffix}"

        match path.suffix:
            case ".csv":
                self._data = pandas.read_csv(path)
            case ".pkl":
                self._data = pandas.read_pickle(path)
            case ".tab" | ".tsv":
                self._data = pandas.read_csv(path, sep="\t")
            case _:
                raise ValueError

        self._x_keys = x_keys
        self._y_keys = y_keys

        self.transform = transform
        self.target_transform = target_transform

        self._x = self._data[self._x_keys].apply(tuple, axis=1)

        if self._y_keys is not None:
            self._y = self._data[self._y_keys].apply(tuple, axis=1)

    def __getitem__(self, index: int) -> Tuple[T, T]:
        x = self._x[index]

        if len(x) == 1:
            x = x[0]

        if self.transform is not None:
            x = self.transform(x)

        if self._y_keys is None:
            return x

        y = self._y[index]

        if len(y) == 1:
            y = y[0]

        if self.target_transform is not None:
            y = self.target_transform(y)

        return x, y

    def __len__(self) -> int:
        return len(self._data)



import beignet
import torch


def test_evaluate_polynomial_cartesian_3d():
    x = torch.rand(3, 5) * 2 - 1

    y = beignet.evaluate_polynomial(
        x,
        torch.tensor([1.0, 2.0, 3.0]),
    )

    a, b, x3 = x
    y1, y2, y3 = y

    torch.testing.assert_close(
        beignet.evaluate_polynomial_cartesian_3d(
            a,
            b,
            x3,
            torch.einsum(
                "i,j,k->ijk",
                torch.tensor([1.0, 2.0, 3.0]),
                torch.tensor([1.0, 2.0, 3.0]),
                torch.tensor([1.0, 2.0, 3.0]),
            ),
        ),
        torch.einsum(
            "i,j,k->ijk",
            y1,
            y2,
            y3,
        ),
    )

    output = beignet.evaluate_polynomial_cartesian_3d(
        torch.ones([2, 3]),
        torch.ones([2, 3]),
        torch.ones([2, 3]),
        torch.einsum(
            "i,j,k->ijk",
            torch.tensor([1.0, 2.0, 3.0]),
            torch.tensor([1.0, 2.0, 3.0]),
            torch.tensor([1.0, 2.0, 3.0]),
        ),
    )

    assert output.shape == (2, 3) * 3



from torch import Tensor

from ._evaluate_probabilists_hermite_polynomial import (
    evaluate_probabilists_hermite_polynomial,
)


def evaluate_probabilists_hermite_polynomial_cartersian_2d(
    x: Tensor,
    y: Tensor,
    c: Tensor,
) -> Tensor:
    for arg in [x, y]:
        c = evaluate_probabilists_hermite_polynomial(arg, c)
    return c



import torch
from torch import Tensor


def physicists_hermite_polynomial_companion(input: Tensor) -> Tensor:
    input = torch.atleast_1d(input)

    if input.shape[0] < 2:
        raise ValueError

    if input.shape[0] == 2:
        return torch.tensor([[-0.5 * input[0] / input[1]]])

    n = input.shape[0] - 1

    output = torch.zeros((n, n), dtype=input.dtype)

    scale = torch.hstack(
        [
            torch.tensor([1.0]),
            1.0 / torch.sqrt(2.0 * torch.arange(n - 1, 0, -1)),
        ],
    )

    scale = torch.cumprod(scale, dim=0)

    scale = torch.flip(scale, dims=[0])

    shp = output.shape

    output = torch.reshape(output, [-1])

    output[1 :: n + 1] = torch.sqrt(0.5 * torch.arange(1, n))
    output[n :: n + 1] = torch.sqrt(0.5 * torch.arange(1, n))

    output = torch.reshape(output, shp)

    output[:, -1] += -scale * input[:-1] / (2.0 * input[-1])

    return output



import beignet
import hypothesis.strategies
import torch.testing


@hypothesis.strategies.composite
def _strategy(function):
    input = torch.randn(
        [
            function(
                hypothesis.strategies.integers(
                    min_value=1,
                    max_value=8,
                ),
            ),
            3,
        ]
    )

    transform = torch.randn([3, 3])

    return (
        (input, transform),
        beignet.apply_transform(
            beignet.apply_transform(input, transform),
            beignet.invert_transform(transform),
        ),
    )


@hypothesis.given(_strategy())
def test_invert_transform(data):
    (input, transform), expected = data

    torch.testing.assert_close(
        input,
        beignet.apply_transform(
            beignet.apply_transform(input, transform),
            beignet.invert_transform(transform),
        ),
    )



from torch import Tensor

from ._evaluate_physicists_hermite_polynomial import (
    evaluate_physicists_hermite_polynomial,
)


def evaluate_physicists_hermite_polynomial_cartesian_2d(
    x: Tensor,
    y: Tensor,
    coefficients: Tensor,
) -> Tensor:
    for point in [x, y]:
        coefficients = evaluate_physicists_hermite_polynomial(
            point,
            coefficients,
        )

    return coefficients



import beignet
import pytest
import torch


def test_differentiate_legendre_polynomial():
    with pytest.raises(TypeError):
        beignet.differentiate_legendre_polynomial(
            torch.tensor([0.0]),
            order=0.5,
        )

    with pytest.raises(ValueError):
        beignet.differentiate_legendre_polynomial(
            torch.tensor([0.0]),
            order=-1,
        )

    for i in range(5):
        torch.testing.assert_close(
            beignet.trim_legendre_polynomial_coefficients(
                beignet.differentiate_legendre_polynomial(
                    torch.tensor([0.0] * i + [1.0]),
                    order=0,
                ),
                tol=0.000001,
            ),
            beignet.trim_legendre_polynomial_coefficients(
                torch.tensor([0.0] * i + [1.0]),
                tol=0.000001,
            ),
        )

    for i in range(5):
        for j in range(2, 5):
            torch.testing.assert_close(
                beignet.trim_legendre_polynomial_coefficients(
                    beignet.differentiate_legendre_polynomial(
                        beignet.integrate_legendre_polynomial(
                            torch.tensor([0.0] * i + [1.0]),
                            order=j,
                        ),
                        order=j,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_legendre_polynomial_coefficients(
                    torch.tensor([0.0] * i + [1.0]),
                    tol=0.000001,
                ),
            )

    for i in range(5):
        for j in range(2, 5):
            torch.testing.assert_close(
                beignet.trim_legendre_polynomial_coefficients(
                    beignet.differentiate_legendre_polynomial(
                        beignet.integrate_legendre_polynomial(
                            torch.tensor([0.0] * i + [1.0]),
                            order=j,
                            scale=2,
                        ),
                        order=j,
                        scale=0.5,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_legendre_polynomial_coefficients(
                    torch.tensor([0.0] * i + [1.0]),
                    tol=0.000001,
                ),
            )

    coefficients = torch.rand(3, 4)

    target = []

    for input in coefficients.T:
        target = [
            *target,
            beignet.differentiate_legendre_polynomial(
                input,
            ),
        ]

    torch.testing.assert_close(
        beignet.differentiate_legendre_polynomial(
            coefficients,
            axis=0,
        ),
        torch.vstack(target).T,
    )

    target = []

    for input in coefficients:
        target = [
            *target,
            beignet.differentiate_legendre_polynomial(
                input,
            ),
        ]

    # torch.testing.assert_close(
    #     beignet.legder(
    #         coefficients,
    #         axis=1,
    #     ),
    #     torch.vstack(target),
    # )

    torch.testing.assert_close(
        beignet.differentiate_legendre_polynomial(
            torch.tensor([1.0, 2.0, 3.0, 4.0]),
            order=4,
        ),
        torch.tensor([0.0]),
    )



import beignet
import torch


def test_divide_chebyshev_polynomial():
    for j in range(5):
        for k in range(5):
            input = torch.tensor([0.0] * j + [1.0])
            other = torch.tensor([0.0] * k + [1.0])

            quotient, remainder = beignet.divide_chebyshev_polynomial(
                beignet.add_chebyshev_polynomial(
                    input,
                    other,
                ),
                input,
            )

            torch.testing.assert_close(
                beignet.trim_chebyshev_polynomial_coefficients(
                    beignet.add_chebyshev_polynomial(
                        beignet.multiply_chebyshev_polynomial(
                            quotient,
                            input,
                        ),
                        remainder,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_chebyshev_polynomial_coefficients(
                    beignet.add_chebyshev_polynomial(
                        input,
                        other,
                    ),
                    tol=0.000001,
                ),
            )



import math

import beignet
import pytest
import torch


def test_evaluate_polynomial_from_roots():
    with pytest.raises(ValueError):
        beignet.evaluate_polynomial_from_roots(
            torch.tensor([1.0]),
            torch.tensor([1.0]),
            tensor=False,
        )

    output = beignet.evaluate_polynomial_from_roots(
        torch.tensor([]),
        torch.tensor([1.0]),
    )

    assert math.prod(output.shape) == 0

    assert output.shape == (0,)

    output = beignet.evaluate_polynomial_from_roots(
        torch.tensor([]),
        torch.tensor([[1.0] * 5]),
    )

    assert math.prod(output.shape) == 0

    assert output.shape == (5, 0)

    torch.testing.assert_close(
        beignet.evaluate_polynomial_from_roots(
            torch.tensor([1.0]),
            torch.tensor([1.0]),
        ),
        torch.tensor([0.0]),
    )

    output = beignet.evaluate_polynomial_from_roots(
        torch.tensor([1.0]),
        torch.ones([3, 3]),
    )

    assert output.shape == (3, 1)

    input = torch.linspace(-1, 1, 50)

    evaluations = []

    for i in range(5):
        evaluations = [*evaluations, input**i]

    for i in range(1, 5):
        target = evaluations[i]

        torch.testing.assert_close(
            beignet.evaluate_polynomial_from_roots(
                input,
                torch.tensor([0.0] * i),
            ),
            target,
        )

    torch.testing.assert_close(
        beignet.evaluate_polynomial_from_roots(
            input,
            torch.tensor([-1.0, 0.0, 1.0]),
        ),
        input * (input - 1.0) * (input + 1.0),
    )

    for i in range(3):
        shape = (2,) * i

        input = torch.zeros(shape)

        output = beignet.evaluate_polynomial_from_roots(
            input,
            torch.tensor([1.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_polynomial_from_roots(
            input,
            torch.tensor([1.0, 0.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_polynomial_from_roots(
            input,
            torch.tensor([1.0, 0.0, 0.0]),
        )

        assert output.shape == shape

    ptest = torch.tensor([15.0, 2.0, -16.0, -2.0, 1.0])

    r = beignet.polynomial_roots(ptest)

    torch.testing.assert_close(
        beignet.evaluate_polynomial(
            input,
            ptest,
        ),
        beignet.evaluate_polynomial_from_roots(
            input,
            r,
        ),
    )

    x = torch.arange(-3, 2)

    r = torch.randint(-5, 5, (3, 5)).to(torch.float64)

    target = torch.empty(r.shape[1:])

    for j in range(math.prod(target.shape)):
        target[j] = beignet.evaluate_polynomial_from_roots(
            x[j],
            r[:, j],
        )

    torch.testing.assert_close(
        beignet.evaluate_polynomial_from_roots(
            x,
            r,
            tensor=False,
        ),
        target,
    )

    x = torch.vstack([x, 2 * x])

    target = torch.empty(r.shape[1:] + x.shape)

    for j in range(r.shape[1]):
        for k in range(x.shape[0]):
            target[j, k, :] = beignet.evaluate_polynomial_from_roots(
                x[k],
                r[:, j],
            )

    torch.testing.assert_close(
        beignet.evaluate_polynomial_from_roots(
            x,
            r,
            tensor=True,
        ),
        target,
    )



import operator

import beignet
import hypothesis.strategies
import torch.testing
from scipy.spatial.transform import Rotation


@hypothesis.strategies.composite
def _strategy(function):
    size = function(
        hypothesis.strategies.integers(
            min_value=16,
            max_value=32,
        ),
    )

    input = Rotation.random(size)
    other = Rotation.random(size)

    return (
        {
            "input": torch.from_numpy(input.as_matrix()),
            "other": torch.from_numpy(other.as_matrix()),
        },
        torch.from_numpy(operator.mul(input, other).as_matrix()),
    )


@hypothesis.given(_strategy())
def test_compose_rotation_matrix(data):
    parameters, expected = data

    torch.testing.assert_close(
        beignet.compose_rotation_matrix(**parameters),
        expected,
    )



import beignet
import torch


def test_physicists_hermite_polynomial_weight():
    torch.testing.assert_close(
        beignet.physicists_hermite_polynomial_weight(torch.linspace(-5, 5, 11)),
        torch.exp(-(torch.linspace(-5, 5, 11) ** 2)),
    )



import math

import beignet
import torch


def test_evaluate_laguerre_polynomial():
    coefficients = [
        torch.tensor([1]) / 1,
        torch.tensor([1, -1]) / 1,
        torch.tensor([2, -4, 1]) / 2,
        torch.tensor([6, -18, 9, -1]) / 6,
        torch.tensor([24, -96, 72, -16, 1]) / 24,
        torch.tensor([120, -600, 600, -200, 25, -1]) / 120,
        torch.tensor([720, -4320, 5400, -2400, 450, -36, 1]) / 720,
    ]

    output = beignet.evaluate_laguerre_polynomial(
        torch.tensor([]),
        torch.tensor([1.0]),
    )

    assert math.prod(output.shape) == 0

    ys = []

    input = torch.linspace(-1, 1, 50)

    for coefficient in coefficients:
        ys = [
            *ys,
            beignet.evaluate_polynomial(
                input,
                coefficient,
            ),
        ]

    for i in range(7):
        torch.testing.assert_close(
            beignet.evaluate_laguerre_polynomial(
                input,
                torch.tensor([0.0] * i + [1.0]),
            ),
            torch.tensor(torch.tensor(ys[i])),
        )

    for index in range(3):
        shape = (2,) * index

        input = torch.zeros(shape)

        output = beignet.evaluate_laguerre_polynomial(
            input,
            torch.tensor([1.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_laguerre_polynomial(
            input,
            torch.tensor([1.0, 0.0]),
        )

        assert output.shape == shape

        output = beignet.evaluate_laguerre_polynomial(
            input,
            torch.tensor([1.0, 0.0, 0.0]),
        )

        assert output.shape == shape



import torch
from torch import Tensor


def differentiate_legendre_polynomial(
    input,
    order=1,
    scale=1,
    axis=0,
) -> Tensor:
    r"""
    Returns the derivative of a polynomial.

    Parameters
    ----------
    input : Tensor
        Polynomial coefficients.

    order : Tensor, optional

    scale : Tensor, optional

    dim : int, default=0

    Returns
    -------
    output : Tensor
        Polynomial coefficients of the derivative.
    """
    if order < 0:
        raise ValueError

    input = torch.atleast_1d(input)

    if order == 0:
        return input

    input = torch.moveaxis(input, axis, 0)

    n = input.shape[0]

    if order >= n:
        input = torch.zeros_like(input[:1])
    else:
        for _ in range(order):
            n = n - 1
            input *= scale
            der = torch.empty((n,) + input.shape[1:], dtype=input.dtype)

            def body(k, der_c, n=n):
                j = n - k

                der, c = der_c

                der[j - 1] = (2 * j - 1) * c[j]

                c[j - 2] += c[j]

                return der, c

            b = n - 2

            x = (der, input)

            y = x

            for index in range(0, b):
                y = body(index, y)

            der, input = y

            if n > 1:
                der[1] = 3 * input[2]

            der[0] = input[1]

            input = der

    input = torch.moveaxis(input, 0, axis)

    return input



from typing import Literal

import torch
from torch import Tensor


def multiply_legendre_polynomial_by_x(
    input: Tensor, mode: Literal["full", "same"] = "full"
) -> Tensor:
    input = torch.atleast_1d(input)

    output = torch.zeros(input.shape[0] + 1, dtype=input.dtype)
    output[1] = input[0]

    for index in range(1, input.shape[0]):
        output[index + 1] = (input[index] * (index + 1)) / (index + index + 1)
        output[index - 1] = output[index - 1] + (input[index] * (index + 0)) / (
            index + index + 1
        )

    if mode == "same":
        output = output[: input.shape[0]]

    return output
